import json

import click
from rich.console import Console
from rich.table import Table

from log10._httpx_utils import _get_time_diff
from log10.feedback.feedback_task import FeedbackTask


# create a cli interface for FeebackTask.create function
@click.command()
@click.option("--name", prompt="Enter feedback task name", help="Name of the task")
@click.option("--task_schema", prompt="Enter feedback task schema", help="Task schema")
@click.option("--instruction", help="Task instruction", default="")
@click.option(
    "--completion_tags_selector",
    help="Completion tags selector",
)
def create_feedback_task(name, task_schema, instruction, completion_tags_selector=None):
    click.echo("Creating feedback task")
    tags = []

    if completion_tags_selector:
        tags = completion_tags_selector.split(",")

    task_schema = json.loads(task_schema)
    task = FeedbackTask().create(
        name=name, task_schema=task_schema, completion_tags_selector=tags, instruction=instruction
    )
    click.echo(f"Use this task_id to add feedback: {task.json()['id']}")


@click.command()
@click.option("--limit", default=25, help="Number of feedback tasks to fetch")
@click.option("--offset", default=0, help="Offset for the feedback tasks")
def list_feedback_task(limit, offset):
    res = FeedbackTask().list(limit=limit, offset=offset)
    feedback_tasks = res.json()

    data_for_table = []

    for task in feedback_tasks["data"]:
        data_for_table.append(
            {
                "id": task["id"],
                "created_at": _get_time_diff(task["created_at"]),
                "name": task["name"],
                "required": task["json_schema"]["required"],
                "instruction": task["instruction"],
            }
        )

    table = Table(title="Feedback Tasks")
    table.add_column("ID", style="dim")
    table.add_column("Created At")
    table.add_column("Name")
    table.add_column("Required")
    table.add_column("Instruction")
    for item in data_for_table:
        required = ", ".join(item["required"]) if item["required"] else ""
        table.add_row(item["id"], item["created_at"], item["name"], required, item["instruction"])

    console = Console()
    console.print(table)


@click.command()
@click.option("--id", help="Get feedback task by ID")
def get_feedback_task(id):
    try:
        res = FeedbackTask().get(id)
    except Exception as e:
        click.echo(f"Error fetching feedback task {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            click.echo(e.response.json()["error"])
        return
    task = json.dumps(res.json())
    console = Console()
    console.print_json(task)



import sys

import openai

from log10.load import log10


log10(openai)

if sys.version_info < (3, 10):
    raise RuntimeError("Python 3.10 or higher is required to run summary feedback llm call.")

try:
    from magentic import SystemMessage, UserMessage, chatprompt
    from magentic.chat_model.openai_chat_model import OpenaiChatModel
    from magentic.chatprompt import escape_braces
except ImportError as error:
    msg = "To use summary feedback llm call, you need to install magentic package. Please install it using 'pip install log10-io[autofeedback_icl]'"
    raise ImportError(msg) from error


# define prompts for tldr dataset
SUMMARY_SYSTEM_PROMPT = """You are an evaluator of summaries of articles on reddit. You are tasked with grading the summaries for accuracy, coherence, coverage and overall.

Coherence
For this axis, answer the question “how coherent is the summary on its own?” A summary is
coherent if, when read by itself, it’s easy to understand and free of English errors. A summary is
not coherent if it’s difficult to understand what the summary is trying to say. Generally, it’s more
important that the summary is understandable than it being free of grammar errors.
Rubric:
Score of 1: The summary is impossible to understand.
Score of 4: The summary has mistakes or confusing phrasing that make it a bit hard to understand.
Score of 7: The summary is perfectly clear.

Accuracy
For this axis, answer the question “does the factual information in the summary accurately match
the post?” A summary is accurate if it doesn’t say things that aren’t in the article, it doesn’t mix up
people, and generally is not misleading. If the summary says anything at all that is not mentioned
in the post or contradicts something in the post, it should be given a maximum score of 5. (If you
are confused about how to use ‘6’, see the FAQ!)
Rubric:
Score of 1: The summary is completely wrong, made up, or exactly contradicts what is written in
the post.
Score of 4: The summary says at least one substantial thing that is not mentioned in the post, or
that contradicts something in the post.
(Score of 5: The summary says anything, no matter how small, that is not mentioned in the post,
or that contradicts something in the post.)
Score of 7: The summary has no incorrect statements or misleading implications.

Coverage
For this axis, answer the question “how well does the summary cover the important information
in the post?” A summary has good coverage if it mentions the main information from the post
that’s important to understand the situation described in the post. A summary has poor coverage if
someone reading only the summary would be missing several important pieces of information
about the situation in the post. A summary with good coverage should also match the purpose of
the original post (e.g. to ask for advice).
Rubric:
Score of 1: The summary contains no information relevant to the post.
Score of 4: The summary is missing at least 1 important piece of information required to understand the situation.
Score of 7: The summary covers all of the important information required to understand the
situation.

Overall quality
For this axis, answer the question “how good is the summary overall at representing the post?”
This can encompass all of the above axes of quality, as well as others you feel are important. If
it’s hard to find ways to make the summary better, give the summary a high score. If there are lots
of different ways the summary can be made better, give the summary a low score.
Rubric:
Score of 1: The summary is terrible.
Score of 4: The summary is an okay representation of the post, but could be significantly improved.
Score of 7: The summary is an excellent representation of the post."""

SUMMARY_USER_MESSAGE = """
Assign scores and write a explanation note for the summary in the test post in json format based on what you think the evaluators would have assigned it.
Do not generate a new summary but just grade the summary that is presented in the last test example
Here is an example format for the final output:
{"note": "This summary is pretty concise but the key points are conveyed here", "axes": {"overall": "6", "accuracy": "6", "coverage": "5", "coherence": "6"}}

Only answer with the scores and note for the final test post and not the example posts.
Remember to not add any additional text beyond the json output
For e.g. don't say things such as "Here is my assessment:" or "Here is the extracted JSON:"
"""

SUMMARY_USER_MESSAGE = escape_braces(SUMMARY_USER_MESSAGE)


@chatprompt(
    SystemMessage(SUMMARY_SYSTEM_PROMPT),
    UserMessage(SUMMARY_USER_MESSAGE),
    UserMessage("Examples: \n{examples}\n\nTest: \n{prompt}"),
    model=OpenaiChatModel("gpt-4-0125-preview", temperature=0.2),
)
def summary_feedback_llm_call(examples, prompt) -> str: ...


def flatten_messages(completion: dict) -> dict:
    request_messages = completion.get("request", {}).get("messages", [])
    if len(request_messages) > 1 and request_messages[1].get("content", ""):
        prompt = request_messages[1].get("content")
    else:
        prompt = ""

    response_choices = completion.get("response", {}).get("choices", [])
    if response_choices and response_choices[0].get("message", {}):
        response = response_choices[0].get("message", {}).get("content", "")
    else:
        response = ""
    return {"prompt": prompt, "response": response}



import asyncio
import json
import logging
import os
import time
import traceback
import uuid
from datetime import datetime, timezone

import httpx
from httpx import Request, Response

from log10.llm import Log10Config
from log10.load import get_log10_session_tags, last_completion_response_var, session_id_var


logger: logging.Logger = logging.getLogger("LOG10")

GRAPHQL_URL = "https://graphql.log10.io/graphql"

LOG10_HTTPX_READ_TIMEOUT = os.environ.get("LOG10_HTTPX_READ_TIMEOUT")
_log10_config = Log10Config()
base_url = _log10_config.url
# Default timeouts for httpx client: connect, read, write, and pool are all 5 seconds.
# We're overriding the read timeout to 10s when LOG10_HTTPX_READ_TIMEOUT is not set.
read_timeout = float(LOG10_HTTPX_READ_TIMEOUT) if LOG10_HTTPX_READ_TIMEOUT else 10.0
timeout = httpx.Timeout(5.0, read=read_timeout)
httpx_client = httpx.Client()
httpx_async_client = httpx.AsyncClient(timeout=timeout)


def _get_time_diff(created_at):
    time = datetime.fromisoformat(created_at)
    now = datetime.now(timezone.utc)
    diff = now - time
    # convert the time difference to human readable format
    if diff.days > 0:
        return f"{diff.days} days ago"
    elif diff.seconds > 3600:
        return f"{diff.seconds//3600} hours ago"
    elif diff.seconds > 60:
        return f"{diff.seconds//60} minutes ago"


def _try_get(url: str, timeout: int = 10) -> httpx.Response:
    httpx_client.headers = {
        "x-log10-token": _log10_config.token,
        "x-log10-organization-id": _log10_config.org_id,
        "Content-Type": "application/json",
    }
    httpx_timeout = httpx.Timeout(timeout)
    try:
        res = httpx_client.get(url, timeout=httpx_timeout)
        res.raise_for_status()
        return res
    except httpx.HTTPError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed anthorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
        raise
    except Exception as e:
        logger.error(f"Error: {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            logger.error(e.response.json()["error"])
        raise


def _try_post_request(url: str, payload: dict = {}) -> httpx.Response:
    headers = {
        "x-log10-token": _log10_config.token,
        "x-log10-organization-id": _log10_config.org_id,
        "Content-Type": "application/json",
    }
    payload["organization_id"] = _log10_config.org_id
    res = None
    try:
        res = httpx_client.post(url, headers=headers, json=payload)
        res.raise_for_status()
        return res
    except httpx.HTTPError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed anthorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
    except Exception as err:
        logger.error(f"Failed to insert in log10: {payload} with error {err}.", exc_info=True)


def _try_post_graphql_request(query: str, variables: dict = {}) -> httpx.Response:
    headers = {"content-type": "application/json", "x-api-token": _log10_config.token}

    payload = {"query": query, "variables": variables}
    res = None
    try:
        res = httpx_client.post(GRAPHQL_URL, headers=headers, json=payload)
        res.raise_for_status()
        return res
    except httpx.HTTPError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed anthorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
    except Exception as err:
        logger.error(f"Failed to make requests to log10 graphQL: {payload} with error {err}")


async def _try_post_graphql_request_async(query: str, variables: dict = {}) -> httpx.Response:
    headers = {"content-type": "application/json", "x-api-token": _log10_config.token}

    payload = {"query": query, "variables": variables}
    res = None
    try:
        res = await httpx_async_client.post(GRAPHQL_URL, headers=headers, json=payload)
        res.raise_for_status()
        return res
    except httpx.HTTPError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed anthorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
    except Exception as err:
        logger.error(f"Failed to make requests to log10 graphQL: {payload} with error {err}")


async def _try_post_request_async(url: str, payload: dict = {}) -> httpx.Response:
    headers = {
        "x-log10-token": _log10_config.token,
        "x-log10-organization-id": _log10_config.org_id,
        "Content-Type": "application/json",
    }
    payload["organization_id"] = _log10_config.org_id
    try:
        res = await httpx_async_client.post(url, headers=headers, json=payload)
        res.raise_for_status()
        return res
    except httpx.ReadTimeout as read_timeout_err:
        logger.error(f"Failed to post request to {url} with {payload} due to a read timeout error: {read_timeout_err}")
    except httpx.HTTPStatusError as http_err:
        if "401" in str(http_err):
            logger.error(
                "Failed authorization. Please verify that LOG10_TOKEN and LOG10_ORG_ID are set correctly and try again."
                + "\nSee https://github.com/log10-io/log10#%EF%B8%8F-setup for details"
            )
        else:
            logger.error(f"Failed with error: {http_err}")
    except Exception as err:
        logger.error(f"Failed to insert in log10: {payload} with error {err}.", exc_info=True)


def format_anthropic_request(request_content) -> str:
    if system_message := request_content.get("system", ""):
        request_content["messages"].insert(0, {"role": "system", "content": system_message})

    for message in request_content.get("messages", []):
        new_content = []
        message_content = message.get("content")

        if isinstance(message_content, list):
            for c in message_content:
                c_type = c.get("type", "")
                if c_type == "image":
                    image_type = c.get("source", {}).get("media_type", "")
                    image_data = c.get("source", {}).get("data", "")
                    new_content.append(
                        {
                            "type": "image_url",
                            "image_url": {"url": f"data:{image_type};base64,{image_data}"},
                        }
                    )
                elif c_type == "tool_use":
                    tool_call = {
                        "id": c.get("id", ""),
                        "type": "function",
                        "function": {
                            "name": c.get("name", ""),
                            "arguments": str(c.get("input", {})),
                        },
                    }
                    message["tool_calls"] = [tool_call]
                    del message["content"]
                elif c_type == "tool_result":
                    content_block_content = c.get("content", "")
                    if isinstance(content_block_content, list):
                        for content in content_block_content:
                            if content.get("type", "") == "text":
                                message["content"] = content.get("text", "")
                else:
                    new_content.append(c)

        if new_content:
            message["content"] = new_content

    new_tools = []
    for tool in request_content.get("tools", []):
        if input_schema := tool.get("input_schema", {}):
            parameters = input_schema
        else:
            parameters = {}
        new_tool = {
            "type": "function",
            "function": {
                "name": tool.get("name", ""),
                "description": tool.get("description", ""),
                "parameters": parameters,
            },
        }
        new_tools.append(new_tool)

    if new_tools:
        request_content["tools"] = new_tools

    return json.dumps(request_content)


def _init_log_row(request: Request):
    start_time = time.time()
    request.started = start_time
    orig_module = ""
    orig_qualname = ""
    request_content_decode = request.content.decode("utf-8")
    host = request.headers.get("host")
    if "openai" in host:
        if "chat" in str(request.url):
            kind = "chat"
            orig_module = "openai.api_resources.chat_completion"
            orig_qualname = "ChatCompletion.create"
        else:
            kind = "completion"
            orig_module = "openai.api_resources.completion"
            orig_qualname = "Completion.create"
    elif "anthropic" in host:
        kind = "chat"
        url_path = request.url
        content_type = request.headers.get("content-type")
        request_content = json.loads(request_content_decode)
        if "/v1/messages" in str(url_path):
            if content_type == "application/json":
                orig_module = "anthropic.resources.messages"
                orig_qualname = "Messages.create"
            else:
                orig_module = "anthropic.resources.messages"
                orig_qualname = "Messages.stream"
        else:
            kind = "completion"
            orig_module = "anthropic.resources.completions"
            orig_qualname = "Completions.create"

        request_content_decode = format_anthropic_request(request_content)

    else:
        logger.debug("Currently logging is only available for async openai and anthropic.")
        return
    log_row = {
        "status": "started",
        "kind": kind,
        "orig_module": orig_module,
        "orig_qualname": orig_qualname,
        "request": request_content_decode,
        "session_id": session_id_var.get(),
    }
    if get_log10_session_tags():
        log_row["tags"] = get_log10_session_tags()

    return log_row


def get_completion_id(request: Request):
    host = request.headers.get("host")
    if "anthropic" in host:
        paths = ["/v1/messages", "/v1/complete"]
        if not any(path in str(request.url) for path in paths):
            logger.debug("Currently logging is only available for anthropic v1/messages and v1/complete.")
            return

    if "openai" in host and "v1/chat/completions" not in str(request.url):
        logger.debug("Currently logging is only available for openai v1/chat/completions.")
        return

    completion_id = str(uuid.uuid4())
    request.headers["x-log10-completion-id"] = completion_id
    last_completion_response_var.set({"completionID": completion_id})
    return completion_id


def patch_response(log_row: dict, llm_response: dict, request: Request):
    current_stack_frame = traceback.extract_stack()
    stacktrace = [
        {
            "file": frame.filename,
            "line": frame.line,
            "lineno": frame.lineno,
            "name": frame.name,
        }
        for frame in current_stack_frame
    ]

    elapsed = time.time() - request.started
    if "anthropic" in request.url.host:
        from anthropic.types.completion import Completion
        from anthropic.types.message import Message

        from log10.anthropic import Anthropic

        if "v1/messages" in str(request.url):
            llm_response = Anthropic.prepare_response(Message(**llm_response))
        elif "v1/complete" in str(request.url):
            prompt = ""
            if request.content:
                content = request.content.decode("utf-8")
                content_json = json.loads(content)
                prompt = content_json.get("prompt", "")

            llm_response = Anthropic.prepare_response(Completion(**llm_response), input_prompt=prompt)
        else:
            logger.debug("Currently logging is only available for anthropic v1/messages and v1/complete.")

    log_row["status"] = "finished"
    log_row["response"] = json.dumps(llm_response)
    log_row["duration"] = int(elapsed * 1000)
    log_row["stacktrace"] = json.dumps(stacktrace)
    if get_log10_session_tags():
        log_row["tags"] = get_log10_session_tags()

    return log_row


class _RequestHooks:
    """
    The class to manage the event hooks for sync requests and initialize the log row.
    The event hooks are:
    - get_completion_id: to generate the completion id
    - log_request: to send the sync request with initial log row to the log10 platform
    """

    def __init__(self):
        logger.debug("LOG10: initializing request hooks")
        self.event_hooks = {
            "request": [self.log_request],
        }
        self.log_row = {}

    def log_request(self, request: httpx.Request):
        logger.debug("LOG10: generating completion id")
        completion_id = get_completion_id(request)
        if not completion_id:
            logger.debug("LOG10: completion id is not generated. Skipping")
            return

        logger.debug("LOG10: sending sync request")
        self.log_row = _init_log_row(request)
        _try_post_request(url=f"{base_url}/api/completions/{completion_id}", payload=self.log_row)


class _AsyncRequestHooks:
    """
    The class to manage the event hooks for async requests and initialize the log row.
    The event hooks are:
    - get_completion_id: to generate the completion id
    - log_request: to send the sync request with initial log row to the log10 platform
    """

    def __init__(self):
        logger.debug("LOG10: initializing async request hooks")
        self.event_hooks = {
            "request": [self.log_request],
        }
        self.log_row = {}

    async def log_request(self, request: httpx.Request):
        logger.debug("LOG10: generating completion id")
        completion_id = get_completion_id(request)
        if not completion_id:
            logger.debug("LOG10: completion id is not generated. Skipping")
            return

        logger.debug("LOG10: sending async request")
        self.log_row = _init_log_row(request)
        asyncio.create_task(
            _try_post_request_async(url=f"{base_url}/api/completions/{completion_id}", payload=self.log_row)
        )


class _LogResponse(Response):
    def __init__(self, *args, **kwargs):
        self.log_row = kwargs.pop("log_row")
        super().__init__(*args, **kwargs)

    def patch_streaming_log(self, duration: int, full_response: str):
        current_stack_frame = traceback.extract_stack()
        stacktrace = [
            {
                "file": frame.filename,
                "line": frame.line,
                "lineno": frame.lineno,
                "name": frame.name,
            }
            for frame in current_stack_frame
        ]

        responses = full_response.split("\n\n")
        response_json = self.parse_response_data(responses)

        self.log_row["response"] = json.dumps(response_json)
        self.log_row["status"] = "finished"
        self.log_row["duration"] = duration
        self.log_row["stacktrace"] = json.dumps(stacktrace)

    def iter_bytes(self, *args, **kwargs):
        full_response = ""
        finished = False
        for chunk in super().iter_bytes(*args, **kwargs):
            full_response += chunk.decode(errors="ignore")
            if self.is_response_end_reached(full_response):
                finished = True
                duration = int(time.time() - self.request.started) * 1000

                completion_id = self.request.headers.get("x-log10-completion-id", "")
                if finished and completion_id:
                    self.patch_streaming_log(duration, full_response)
                    _try_post_request(url=f"{base_url}/api/completions/{completion_id}", payload=self.log_row)
            yield chunk

    async def aiter_bytes(self, *args, **kwargs):
        full_response = ""
        finished = False
        async for chunk in super().aiter_bytes(*args, **kwargs):
            full_response += chunk.decode(errors="ignore")
            if self.is_response_end_reached(full_response):
                finished = True
                duration = int(time.time() - self.request.started) * 1000

                completion_id = self.request.headers.get("x-log10-completion-id", "")
                if finished and completion_id:
                    self.patch_streaming_log(duration, full_response)
                    asyncio.create_task(
                        _try_post_request_async(
                            url=f"{base_url}/api/completions/{completion_id}", payload=self.log_row
                        )
                    )
            yield chunk

    def is_response_end_reached(self, text: str):
        host = self.request.headers.get("host")
        if "anthropic" in host:
            return self.is_anthropic_response_end_reached(text)
        elif "openai" in host:
            return self.is_openai_response_end_reached(text)
        else:
            logger.debug("Currently logging is only available for async openai and anthropic.")
            return False

    def is_anthropic_response_end_reached(self, text: str):
        return "event: message_stop" in text

    def is_openai_response_end_reached(self, text: str):
        return "data: [DONE]" in text

    def parse_anthropic_responses(self, responses: list[str]):
        message_id = ""
        model = ""
        finish_reason = None
        full_content = ""
        input_tokens = 0
        output_tokens = 0
        tool_calls = []
        tool_call = {}
        arguments = ""

        for r in responses:
            if not r:
                break

            data_index = r.find("data:")
            r_json = json.loads(r[data_index + len("data:") :])

            ### anthropic first data contains
            ### {"event":"message_start","data":{"message":{"role":"user","content":"Hello, how are you today?"}}}
            type = r_json["type"]
            if type == "message_start":
                message = r_json.get("message", {})
                message_id = message.get("id", "")
                model = message.get("model", "")
                input_tokens = message.get("usage", {}).get("input_tokens", 0)
            elif type == "content_block_start":
                content_block = r_json.get("content_block", {})
                content_block_type = content_block.get("type", "")
                if content_block_type == "tool_use":
                    tool_call = {
                        "id": content_block.get("id", ""),
                        "type": "function",
                        "function": {"name": content_block.get("name", ""), "arguments": ""},
                    }
                if content_block_type == "text":
                    full_content += content_block.get("text", "")
            elif type == "content_block_delta":
                delta = r_json.get("delta", {})
                if (delta_text := delta.get("text")) is not None:
                    full_content += delta_text

                if (delta_partial_json := delta.get("partial_json")) is not None:
                    ## If there is a content of chain of thoughts,
                    ## then the next one should be append to the tool_call
                    arguments += delta_partial_json
            elif type == "message_delta":
                finish_reason = r_json.get("delta", {}).get("stop_reason", "")
                output_tokens = r_json.get("usage", {}).get("output_tokens", 0)
            elif type == "content_block_stop" or type == "message_stop":
                if tool_call:
                    tool_call["function"]["arguments"] = arguments
                    tool_calls.append(tool_call)
                    arguments = ""
                    tool_call = {}

        response = {
            "id": message_id,
            "object": "chat",
            "model": model,
            "choices": [
                {
                    "index": 0,
                    "finish_reason": finish_reason,
                }
            ],
            "usage": {
                "prompt_tokens": input_tokens,
                "completion_tokens": output_tokens,
                "total_tokens": input_tokens + output_tokens,
            },
        }

        response_json = response.copy()
        message = {
            "role": "assistant",
        }

        if full_content:
            message["content"] = full_content

        if tool_calls:
            message["tool_calls"] = tool_calls

        response_json["choices"][0]["message"] = message
        return response_json

    def parse_openai_responses(self, responses: list[str]):
        r_json = {}
        tool_calls = []
        full_content = ""
        function_name = ""
        full_argument = ""
        finish_reason = ""

        for r in responses:
            if self.is_openai_response_end_reached(r):
                break

            # loading the substring of response text after 'data: '.
            # example: 'data: {"choices":[{"text":"Hello, how can I help you today?"}]}'
            data_index = r.find("data:")
            r_json = json.loads(r[data_index + len("data:") :])

            if r_json.get("choices", []):
                if delta := r_json["choices"][0].get("delta", {}):
                    # Delta may have content
                    # delta: { "content": " "}
                    if content := delta.get("content", ""):
                        full_content += content

                    # May be a function call, and have to reconstruct the arguments
                    if function_call := delta.get("function_call", {}):
                        # May be function name
                        if function_call_name := function_call.get("name", ""):
                            function_name = function_call_name
                        # May be function arguments
                        if arguments := function_call.get("arguments", ""):
                            full_argument += arguments

                    if tc := delta.get("tool_calls", []):
                        if tc[0].get("id", ""):
                            tool_calls.append(tc[0])
                        elif tc[0].get("function", {}).get("arguments", ""):
                            idx = tc[0].get("index")
                            tool_calls[idx]["function"]["arguments"] += tc[0]["function"]["arguments"]

                if fr := r_json["choices"][0].get("finish_reason", ""):
                    finish_reason = fr

        r_json["object"] = "chat.completion"
        # r_json is the last response before "data: [DONE]"
        response_json = r_json.copy()

        if not response_json.get("choices"):
            response_json["choices"] = [{"index": 0}]

        if response_json.get("choices", []):
            # It will only set finish_reason for openai version > 1.26.0
            if finish_reason:
                response_json["choices"][0]["finish_reason"] = finish_reason

            if full_content:
                response_json["choices"][0]["message"] = {
                    "role": "assistant",
                    "content": full_content,
                }
            elif tool_calls:
                response_json["choices"][0]["message"] = {
                    "content": None,
                    "role": "assistant",
                    "tool_calls": tool_calls,
                }
            elif function_name and full_argument:
                # function is deprecated in openai api
                response_json["choices"][0]["function_call"] = {
                    "name": function_name,
                    "arguments": full_argument,
                }

        return response_json

    def parse_response_data(self, responses: list[str]):
        host = self.request.headers.get("host")
        if "openai" in host:
            return self.parse_openai_responses(responses)
        elif "anthropic" in host:
            return self.parse_anthropic_responses(responses)
        else:
            logger.debug("Currently logging is only available for async openai and anthropic.")
            return None


class _LogTransport(httpx.BaseTransport):
    def __init__(self, transport: httpx.BaseTransport, request_hooks: _RequestHooks):
        self.transport = transport
        self.request_hooks = request_hooks

    def handle_request(self, request: Request) -> Response:
        try:
            response = self.transport.handle_request(request)
        except Exception as e:
            logger.warning(f"Failed to send request: {e}")
            return

        if response.status_code >= 400:
            logger.warning(f"HTTP error occurred: {response.status_code}")
            return

        completion_id = request.headers.get("x-log10-completion-id", "")
        if not completion_id:
            return response

        if response.headers.get("content-type").startswith("application/json"):
            response.read()
            llm_response = response.json()
            log_row = patch_response(self.request_hooks.log_row, llm_response, request)
            _try_post_request(url=f"{base_url}/api/completions/{completion_id}", payload=log_row)
            return response
        elif response.headers.get("content-type").startswith("text/event-stream"):
            return _LogResponse(
                status_code=response.status_code,
                headers=response.headers,
                stream=response.stream,
                extensions=response.extensions,
                request=request,
                log_row=self.request_hooks.log_row,
            )

        # In case of an error, get out of the way
        return response


class _AsyncLogTransport(httpx.AsyncBaseTransport):
    def __init__(self, transport: httpx.AsyncBaseTransport, async_request_hooks: _AsyncRequestHooks):
        self.transport = transport
        self.async_request_hooks = async_request_hooks

    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        try:
            response = await self.transport.handle_async_request(request)
        except Exception as e:
            logger.warning(f"Failed to send request: {e}")
            return

        if response.status_code >= 400:
            logger.warning(f"HTTP error occurred: {response.status_code}")
            return

        completion_id = request.headers.get("x-log10-completion-id", "")
        if not completion_id:
            return response

        if response.headers.get("content-type").startswith("application/json"):
            await response.aread()
            llm_response = response.json()
            log_row = patch_response(self.async_request_hooks.log_row, llm_response, request)
            asyncio.create_task(
                _try_post_request_async(url=f"{base_url}/api/completions/{completion_id}", payload=log_row)
            )
            return response
        elif response.headers.get("content-type").startswith("text/event-stream"):
            return _LogResponse(
                status_code=response.status_code,
                headers=response.headers,
                stream=response.stream,
                extensions=response.extensions,
                request=request,
                log_row=self.async_request_hooks.log_row,
            )

        # In case of an error, get out of the way
        return response


class InitPatcher:
    def __init__(self, module, class_names: list[str]):
        logger.debug("LOG10: initializing patcher")

        allowed_modules = ["openai", "anthropic"]
        if not any(allowed_module in module.__name__ for allowed_module in allowed_modules):
            raise ValueError("Only openai and anthropic modules are allowed.")

        self.module = module
        if len(class_names) > 2:
            raise ValueError("Only two class names (sync and async) are allowed")

        self.async_class_name = None
        self.sync_class_name = None

        for class_name in class_names:
            if class_name.startswith("Async"):
                self.async_class_name = class_name
                self.async_origin_init = getattr(module, self.async_class_name).__init__
            else:
                self.sync_class_name = class_name
                self.origin_init = getattr(module, self.sync_class_name).__init__

        self._patch_init()

    def _patch_init(self):
        def new_init(instance, *args, **kwargs):
            logger.debug(f"LOG10: patching {self.sync_class_name}.__init__")

            request_hooks = _RequestHooks()
            httpx_client = httpx.Client(
                event_hooks=request_hooks.event_hooks,
                transport=_LogTransport(httpx.HTTPTransport(), request_hooks),
            )
            kwargs["http_client"] = httpx_client
            self.origin_init(instance, *args, **kwargs)

        def async_new_init(instance, *args, **kwargs):
            logger.debug(f"LOG10: patching {self.async_class_name}.__init__")

            async_request_hooks = _AsyncRequestHooks()
            async_httpx_client = httpx.AsyncClient(
                event_hooks=async_request_hooks.event_hooks,
                transport=_AsyncLogTransport(httpx.AsyncHTTPTransport(), async_request_hooks),
            )
            kwargs["http_client"] = async_httpx_client
            self.async_origin_init(instance, *args, **kwargs)

        # Patch the asynchronous class __init__
        if self.async_class_name:
            async_class = getattr(self.module, self.async_class_name)
            async_class.__init__ = async_new_init

        # Patch the synchronous class __init__ if provided
        if self.sync_class_name:
            sync_class = getattr(self.module, self.sync_class_name)
            sync_class.__init__ = new_init


async def finalize():
    pending = asyncio.all_tasks()
    pending.remove(asyncio.current_task())
    await asyncio.gather(*pending)



import json
import logging
import random
from types import FunctionType

import httpx

from log10._httpx_utils import _try_post_graphql_request
from log10.completions.completions import _get_completion
from log10.feedback.feedback import _get_feedback_list
from log10.llm import Log10Config
from log10.load import log10_session


try:
    from log10.feedback._summary_feedback_utils import flatten_messages, summary_feedback_llm_call

    Magentic_imported = True
except ImportError:
    Magentic_imported = False


logger = logging.getLogger("LOG10")
logger.setLevel(logging.INFO)

_log10_config = Log10Config()


class AutoFeedbackICL:
    """
    Generate feedback with in context learning (ICL) based on existing feedback.
    """

    _examples: list[dict] = []
    _predict_func: FunctionType = None

    def __init__(
        self,
        task_id: str,
        num_samples: int = 5,
        predict_func: FunctionType = summary_feedback_llm_call if Magentic_imported else None,
    ):
        if not Magentic_imported:
            raise ImportError(
                "Log10 feedback predict requires magentic package. Please install using 'pip install log10-io[autofeedback_icl]'"
            )
        self.num_samples = num_samples
        self.task_id = task_id
        self._predict_func = predict_func

    def _get_examples(self):
        logger.info(f"Getting {self.num_samples} feedback for task {self.task_id}")
        feedback_data = _get_feedback_list(offset=0, limit="", task_id=self.task_id)
        assert feedback_data, f"No feedback found for task {self.task_id}."
        assert (
            len(feedback_data) >= self.num_samples
        ), f"Insufficient feedback for task {self.task_id}, found {len(feedback_data)} feedback. Sample size {self.num_samples}."
        sampled_feedback = random.sample(feedback_data, self.num_samples)
        few_shot_examples = []
        for fb in sampled_feedback:
            feedback_values = fb["json_values"]
            completion_id = fb["matched_completion_ids"][0]
            try:
                res = _get_completion(completion_id)
            except Exception as e:
                print(e)
                continue
            completion = res.json()["data"]
            fm = flatten_messages(completion)

            few_shot_examples.append(
                {
                    "completion_id": completion_id,
                    "prompt": fm["prompt"],
                    "response": fm["response"],
                    "feedback": json.dumps(feedback_values),
                }
            )
        logger.info(f"Sampled completion ids: {[d['completion_id'] for d in few_shot_examples]}")
        return few_shot_examples

    def predict(self, text: str = None, completion_id: str = None) -> str:
        if not self._examples:
            self._examples = self._get_examples()

        # Here assumps the completion is summary, prompt is article, response is summary
        if completion_id and not text:
            completion = _get_completion(completion_id)
            pr = flatten_messages(completion.json()["data"])
            text = json.dumps(pr)

        logger.info(f"{text=}")

        predict_func_name = self._predict_func.__name__
        logger.info(f"Using predict llm_call: {predict_func_name}")
        with log10_session(tags=["autofeedback_icl", predict_func_name]):
            ret = self._predict_func(examples="\n".join([str(d) for d in self._examples]), prompt=text)
        return ret


def get_autofeedback(completion_id: str) -> httpx.Response:
    query = """
    query OrganizationCompletion($orgId: String!, $id: String!) {
        organization(id: $orgId) {
            completion(id: $id) {
                id
                autoFeedback {
                    id
                    status
                    jsonValues
                    comment
                }
            }
        }
    }
    """

    variables = {"orgId": _log10_config.org_id, "id": completion_id}

    response = _try_post_graphql_request(query, variables)

    if response is None:
        logger.error(f"Failed to get auto feedback for completion {completion_id}")
        return None

    if response.status_code == 200:
        return response.json()
    else:
        response.raise_for_status()



import asyncio
import contextvars
import functools
import inspect
import json
import logging
import os
import queue
import threading
import time
import traceback
import uuid
from contextlib import contextmanager
from copy import deepcopy
from importlib.metadata import version

import backoff
import requests
from dotenv import load_dotenv
from packaging.version import parse


load_dotenv()

logging.basicConfig(
    format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger: logging.Logger = logging.getLogger("LOG10")

url = os.environ.get("LOG10_URL")
token = os.environ.get("LOG10_TOKEN")
org_id = os.environ.get("LOG10_ORG_ID")


# log10, bigquery
target_service = os.environ.get("LOG10_DATA_STORE", "log10")

if target_service == "bigquery":
    from log10.bigquery import initialize_bigquery

    bigquery_client, bigquery_table = initialize_bigquery()
    from datetime import datetime, timezone
elif target_service is None:
    target_service = "log10"  # default to log10


def is_openai_v1() -> bool:
    """Return whether OpenAI API is v1 or more."""
    _version = parse(version("openai"))
    return _version.major >= 1


def func_with_backoff(func, *args, **kwargs):
    if func.__module__ != "openai" or is_openai_v1():
        return func(*args, **kwargs)

    import openai

    retry_errors = (
        openai.error.APIConnectionError,
        openai.error.APIError,
        openai.error.RateLimitError,
        openai.error.ServiceUnavailableError,
        openai.error.Timeout,
    )

    @backoff.on_exception(backoff.expo, retry_errors)
    def _func_with_backoff(func, *args, **kwargs):
        return func(*args, **kwargs)

    return _func_with_backoff(func, *args, **kwargs)


# todo: should we do backoff as well?
def post_request(url: str, json_payload: dict = {}) -> requests.Response:
    headers = {"x-log10-token": token, "Content-Type": "application/json"}
    json_payload["organization_id"] = org_id
    try:
        timeout = int(os.environ.get("LOG10_REQUESTS_TIMEOUT", 10))
        assert isinstance(timeout, int)
        res = requests.post(url, headers=headers, json=json_payload, timeout=timeout)
        # raise_for_status() will raise an exception if the status is 4xx, 5xxx
        res.raise_for_status()

        logger.debug(f"HTTP request: POST {url} {res.status_code}\n{json.dumps(json_payload, indent=4)}")
        return res
    except requests.Timeout:
        logger.error("HTTP request: POST Timeout")
        raise
    except requests.ConnectionError:
        logger.error("HTTP request: POST Connection Error")
        raise
    except requests.HTTPError as e:
        logger.error(f"HTTP request: POST HTTP Error - {e}")
        raise
    except requests.RequestException as e:
        logger.error(f"HTTP request: POST Request Exception - {e}")
        raise


post_session_request = functools.partial(post_request, url + "/api/sessions", {})


def get_session_id():
    id = str(uuid.uuid4())
    logger.debug(f"Session ID: {id}")
    return id


#
# Context variables
#
session_id_var = contextvars.ContextVar("session_id", default=get_session_id())
last_completion_response_var = contextvars.ContextVar("last_completion_response", default=None)
tags_var = contextvars.ContextVar("tags", default=[])


def get_log10_session_tags():
    return tags_var.get()


class log10_session:
    def __init__(self, tags=None):
        self.tags = tags

    def __enter__(self):
        self.session_id_token = session_id_var.set(get_session_id())
        self.last_completion_response_token = last_completion_response_var.set(None)
        self.tags_token = tags_var.set(self.tags)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        session_id_var.reset(self.session_id_token)
        last_completion_response_var.reset(self.last_completion_response_token)
        tags_var.reset(self.tags_token)
        return

    async def __aenter__(self):
        self.session_id_token = session_id_var.set(get_session_id())
        self.last_completion_response_token = last_completion_response_var.set(None)
        self.tags_token = tags_var.set(self.tags)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        session_id_var.reset(self.session_id_token)
        last_completion_response_var.reset(self.last_completion_response_token)
        tags_var.reset(self.tags_token)
        return

    def last_completion_url(self):
        if last_completion_response_var.get() is None:
            return None
        response = last_completion_response_var.get()

        # organizationSlug will not be returned from httpx hook
        if not response.get("organizationSlug", ""):
            return None
        return f'{url}/app/{response["organizationSlug"]}/completions/{response["completionID"]}'

    def last_completion_id(self):
        if last_completion_response_var.get() is None:
            return None
        response = last_completion_response_var.get()
        return response["completionID"]


@contextmanager
def timed_block(block_name):
    if DEBUG:
        start_time = time.perf_counter()
        try:
            yield
        finally:
            elapsed_time = time.perf_counter() - start_time
            logger.debug(f"TIMED BLOCK - {block_name} took {elapsed_time:.6f} seconds to execute.")
    else:
        yield


def log_url(res, completionID):
    output = res.json()
    organizationSlug = output["organizationSlug"]
    full_url = url + "/app/" + organizationSlug + "/completions/" + completionID
    logger.debug(f"Completion URL: {full_url}")


async def log_async(completion_url, log_row):
    res = None
    try:
        res = post_request(completion_url)
        completionID = res.json().get("completionID", None)
        organizationSlug = res.json().get("organizationSlug", None)

        if completionID is None:
            logging.warn("LOG10: failed to get completionID from log10. Skipping log.")
            return None

        if DEBUG:
            log_url(res, completionID)

        if target_service == "log10":
            try:
                _url = f"{completion_url}/{completionID}"
                res = post_request(_url, log_row)
            except Exception as e:
                logging.warn(f"LOG10: failed to log: {e}. Skipping")
                return None

        elif target_service == "bigquery":
            pass
            # NOTE: We only save on request finalization.

    except Exception as e:
        logging.warn(f"LOG10: failed to log: {e}. Skipping")
        return None

    return {"completionID": completionID, "organizationSlug": organizationSlug}


def run_async_in_thread(completion_url, log_row, result_queue):
    result = asyncio.run(log_async(completion_url=completion_url, log_row=log_row))
    result_queue.put(result)


def log_sync(completion_url, log_row):
    completionID = None

    try:
        res = post_request(completion_url)
        last_completion_response_var.set(res.json())
        completionID = res.json().get("completionID", None)

        if completionID is None:
            logging.warn("LOG10: failed to get completionID from log10. Skipping log.")
            return None

        if DEBUG:
            log_url(res, completionID)
        _url = f"{completion_url}/{completionID}"
        res = post_request(_url, log_row)
    except Exception as e:
        logging.warn(f"LOG10: failed to get completionID from log10: {e}")
        return None

    return completionID


class StreamingResponseWrapper:
    """
    Wraps a streaming response object to log the final result and duration to log10.

    Openai V1 example:
    Example:
        >>> from log10.load import OpenAI
        >>> client = OpenAI()
        >>> response = client.chat.completions.create(
        >>>     model="gpt-3.5-turbo",
        >>>     messages=[{"role": "user", "content": "Count to 200"}],
        >>>     temperature=0,
        >>>     stream=True,
        >>> )
        >>> for chunk in response:
        >>>     content = chunk.choices[0].delta.content
        >>>     if content:
        >>>         print(content, end="", flush=True)
        >>> print("")
    """

    def __init__(self, completion_url, completionID, response, partial_log_row):
        self.completionID = completionID
        self.completion_url = completion_url
        self.partial_log_row = partial_log_row
        self.response = response
        self.final_result = ""  # Store the final result
        self.function_name = ""
        self.function_arguments = ""
        self.start_time = time.perf_counter()
        self.gpt_id = None
        self.model = None
        self.finish_reason = None
        self.usage = None
        self.tool_calls = []

    def __iter__(self):
        return self

    def __next__(self):
        try:
            chunk = next(self.response)
            if hasattr(chunk.choices[0].delta, "content") and chunk.choices[0].delta.content is not None:
                # Here you can intercept and modify content if needed
                content = chunk.choices[0].delta.content
                self.final_result += content  # Save the content
                # Yield the original or modified content

                self.model = chunk.model
                self.gpt_id = chunk.id

                # for mistral stream
                if chunk.choices[0].finish_reason:
                    self.finish_reason = chunk.choices[0].finish_reason

                # for mistral stream
                if getattr(chunk, "usage", None):
                    self.usage = chunk.usage
            elif chunk.choices[0].delta.function_call:
                arguments = chunk.choices[0].delta.function_call.arguments
                self.function_arguments += arguments
                if not self.function_name and chunk.choices[0].delta.function_call.name:
                    self.function_name = chunk.choices[0].delta.function_call.name
            elif tc := chunk.choices[0].delta.tool_calls:
                # self.tool_calls is a list
                if tc[0].id:
                    self.tool_calls.append(tc[0].dict())
                elif tc[0].function.arguments:
                    self.tool_calls[tc[0].index]["function"]["arguments"] += tc[0].function.arguments
            elif chunk.choices[0].finish_reason:
                self.finish_reason = chunk.choices[0].finish_reason

                # in Magentic stream, this is reached instead of StopIteration
                if self.finish_reason == "stop":
                    raise StopIteration

            return chunk
        except StopIteration as se:
            # Log the final result
            # Create fake response for openai format.
            if self.final_result:
                response = {
                    "id": self.gpt_id,
                    "object": "completion",
                    "model": self.model,
                    "choices": [
                        {
                            "index": 0,
                            "finish_reason": self.finish_reason,
                            "message": {
                                "role": "assistant",
                                "content": self.final_result,
                            },
                        }
                    ],
                }
            elif self.function_arguments:
                response = {
                    "id": self.gpt_id,
                    "object": "completion",
                    "model": self.model,
                    "choices": [
                        {
                            "index": 0,
                            "finish_reason": self.finish_reason,
                            "function_call": {
                                "name": self.function_name,
                                "arguments": self.function_arguments,
                            },
                        }
                    ],
                }
            elif self.tool_calls:
                response = {
                    "id": self.gpt_id,
                    "object": "completion",
                    "model": self.model,
                    "choices": [
                        {
                            "index": 0,
                            "finish_reason": self.finish_reason,
                            "message": {
                                "content": "",
                                "role": "assistant",
                                "tool_calls": self.tool_calls,
                            },
                        }
                    ],
                }
            if self.usage:
                response["usage"] = self.usage.dict()
            self.partial_log_row["response"] = json.dumps(response)
            self.partial_log_row["duration"] = int((time.perf_counter() - self.start_time) * 1000)

            try:
                _url = f"{self.completion_url}/{self.completionID}"
                res = post_request(_url, self.partial_log_row)
                if res.status_code != 200:
                    logger.error(f"LOG10: failed to insert in log10: {self.partial_log_row} with error {res.text}")
            except Exception as e:
                traceback.print_tb(e.__traceback__)
                logging.warn(f"LOG10: failed to log: {e}. Skipping")

            raise se


# Filter large images from messages, and replace with a text message saying "Image too large to display"
def filter_large_images(messages):
    for message in messages:
        # Content may be an array of fragments, of text and images.
        # If not, it's a single fragment.
        if isinstance(message.get("content"), list):
            new_content = []
            for fragment in message.get("content", ""):
                if fragment.get("type") == "image_url":
                    # If image is more than 4MB, replace with a text message
                    url = fragment.get("image_url", {}).get("url", "")
                    if url.startswith("data:image"):
                        if len(url) > 4e6:
                            new_content.append(
                                {
                                    "type": "text",
                                    "text": "Image too large to capture",
                                }
                            )
                        else:
                            new_content.append(fragment)
                    else:
                        new_content.append(fragment)
                else:
                    new_content.append(fragment)
            message["content"] = new_content

    return messages


def flatten_messages(messages):
    flat_messages = []
    for message in messages:
        if isinstance(message, dict):
            flat_messages.append(message)
        else:
            flat_messages.append(message.model_dump())
    return flat_messages


def flatten_response(response):
    if "choices" in response:
        # May have to flatten, if not a dictionary
        if not isinstance(response.choices[0].message, dict):
            response.choices[0].message = response.choices[0].message.model_dump()

    return response


def _get_stack_trace():
    current_stack_frame = traceback.extract_stack()
    return [
        {
            "file": frame.filename,
            "line": frame.line,
            "lineno": frame.lineno,
            "name": frame.name,
        }
        for frame in current_stack_frame
    ]


def _init_log_row(func, *args, **kwargs):
    kwargs_copy = deepcopy(kwargs)

    log_row = {
        "status": "started",
        "orig_module": func.__module__,
        "orig_qualname": func.__qualname__,
        "stacktrace": json.dumps(_get_stack_trace()),
        "session_id": session_id_var.get(),
        "tags": tags_var.get(),
    }

    # in case the usage of load(openai) and langchain.ChatOpenAI
    if "api_key" in kwargs_copy:
        kwargs_copy.pop("api_key")

    # We may have to flatten messages from their ChatCompletionMessage with nested ChatCompletionMessageToolCall to json serializable format
    # Rewrite in-place
    if "messages" in kwargs_copy:
        kwargs_copy["messages"] = filter_large_images(flatten_messages(kwargs_copy["messages"]))

    # kind and request are set based on the module and qualname
    # request is based on openai schema
    if "anthropic" in func.__module__:
        logger.debug("Anthropic calls are patched via httpx client, should not reach here.")
    elif "vertexai" in func.__module__:
        if func.__name__ == "_send_message":
            # get model name save in ChatSession instance
            log_row["kind"] = "chat"
            chat_session_instance = inspect.currentframe().f_back.f_back.f_locals["self"]
            model_name = chat_session_instance._model._model_name.split("/")[-1]

            # TODO how to handle chat history
            # chat_history = chat_session_instance.history
            kwargs_copy.update(
                {
                    "model": model_name,
                    "messages": [{"role": "user", "content": kwargs_copy["content"]}],
                }
            )
            if kwargs_copy.get("generation_config"):
                for key, value in kwargs_copy["generation_config"].to_dict().items():
                    if key == "max_output_tokens":
                        kwargs_copy["max_tokens"] = value
                    else:
                        kwargs_copy[key] = value
                kwargs_copy.pop("generation_config")
    elif "lamini" in func.__module__:
        log_row["kind"] = "chat"
        kwargs_copy.update(
            {
                "model": args[1]["model_name"],
                "messages": [{"role": "user", "content": args[1]["prompt"]}],
            }
        )
    elif "mistralai" in func.__module__:
        log_row["kind"] = "chat"
    elif "openai" in func.__module__:
        from openai._utils._utils import strip_not_given

        kwargs_copy = strip_not_given(kwargs_copy)
        kind = "chat" if "chat" in func.__module__ else "completion"
        log_row["kind"] = kind
    elif "google.generativeai" in func.__module__:
        if func.__name__ == "send_message":
            log_row["kind"] = "chat"
            history_messages = []
            if system_instruction := args[0].model._system_instruction:
                history_messages.append({"role": "system", "content": system_instruction.parts[0].text})
            if history := args[0].history:
                for m in history:
                    role = "assistant" if m.role == "model" else "user"
                    content = m.parts[0].text
                    history_messages.append({"role": role, "content": content})

            history_messages.append({"role": "user", "content": args[1]})
            kwargs_copy.update(
                {
                    "model": args[0].model.model_name.split("/")[-1],
                    "messages": history_messages,
                }
            )
            if kwargs_copy.get("generation_config") and hasattr(kwargs_copy["generation_config"], "__dict__"):
                for key, value in kwargs_copy["generation_config"].__dict__.items():
                    if not value:
                        continue
                    if key == "max_output_tokens":
                        kwargs_copy["max_tokens"] = value
                    elif key == "stop_sequences":
                        kwargs_copy["stop"] = value
                    elif key in ["temperature", "top_p", "top_k"]:
                        kwargs_copy[key] = value
                kwargs_copy.pop("generation_config")

    log_row["request"] = json.dumps(kwargs_copy)

    return log_row


def intercepting_decorator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        completion_url = url + "/api/completions"
        output = None
        result_queue = queue.Queue()

        try:
            log_row = _init_log_row(func, *args, **kwargs)

            with timed_block(sync_log_text + " call duration"):
                if USE_ASYNC:
                    threading.Thread(
                        target=run_async_in_thread,
                        kwargs={
                            "completion_url": completion_url,
                            "log_row": log_row,
                            "result_queue": result_queue,
                        },
                    ).start()
                else:
                    completionID = log_sync(completion_url=completion_url, log_row=log_row)

                    if completionID is None:
                        logging.warn("LOG10: failed to get completionID from log10. Skipping log.")
                        func_with_backoff(func, *args, **kwargs)
                        return

            start_time = time.perf_counter()
            output = func_with_backoff(func, *args, **kwargs)
            duration = time.perf_counter() - start_time
            logger.debug(f"TIMED BLOCK - LLM call duration: {duration}")
        except Exception as e:
            if USE_ASYNC:
                with timed_block("extra time spent waiting for log10 call"):
                    while result_queue.empty():
                        pass
                    result = result_queue.get()
                    completionID = result["completionID"]
                    last_completion_response_var.set(result)

            if completionID is None:
                logger.warning(f"LOG10: failed to get completionID from log10: {e}. Skipping log.")
                return

            logger.debug(f"LOG10: failed - {e}")
            # todo: change with openai v1 update
            if type(e).__name__ == "InvalidRequestError" and "This model's maximum context length" in str(e):
                failure_kind = "ContextWindowExceedError"
            else:
                failure_kind = type(e).__name__
            failure_reason = str(e)
            log_row["status"] = "failed"
            log_row["failure_kind"] = failure_kind
            log_row["failure_reason"] = failure_reason
            try:
                res = post_request(completion_url + "/" + completionID, log_row)
            except Exception as le:
                logger.warning(f"LOG10: failed to log: {le}. Skipping, but raising LLM error.")
            raise e
        else:
            # finished with no exceptions
            if USE_ASYNC:
                with timed_block("extra time spent waiting for log10 call"):
                    while result_queue.empty():
                        pass
                    result = result_queue.get()

                    if result is None:
                        return output

                    completionID = result.get("completionID")
                    last_completion_response_var.set(result)

            with timed_block("result call duration (sync)"):
                response = output
                # Adjust the Anthropic output to match OAI completion output
                if "anthropic" in func.__module__:
                    logger.debug("Anthropic calls are patched via httpx client, should not reach here.")

                elif "vertexai" in func.__module__:
                    response = output
                    reason = response.candidates[0].finish_reason.name
                    ret_response = {
                        "id": str(uuid.uuid4()),
                        "object": "chat.completion",
                        "choices": [
                            {
                                "index": 0,
                                "finish_reason": str(reason).lower(),
                                "message": {
                                    "role": "assistant",
                                    "content": response.text,
                                },
                            }
                        ],
                    }
                    response_dict = response.to_dict()
                    tokens_usage = {
                        "prompt_tokens": response_dict["usage_metadata"]["prompt_token_count"],
                        "completion_tokens": response_dict["usage_metadata"]["candidates_token_count"],
                        "total_tokens": response_dict["usage_metadata"]["total_token_count"],
                    }
                    ret_response["usage"] = tokens_usage
                    response = ret_response
                elif "google.generativeai" in func.__module__:
                    ret_response = {
                        "id": str(uuid.uuid4()),
                        "object": "chat.completion",
                        "choices": [
                            {
                                "index": 0,
                                "finish_reason": str(output.candidates[0].finish_reason.name).lower(),
                                "message": {
                                    "role": "assistant",
                                    "content": output.text,
                                },
                            }
                        ],
                        "usage": {
                            "prompt_tokens": 0,
                            "completion_tokens": 0,
                            "total_tokens": 0,
                        },
                    }
                    response = ret_response
                elif "openai" in func.__module__:
                    if type(output).__name__ == "Stream":
                        log_row["response"] = response
                        log_row["status"] = "finished"
                        return StreamingResponseWrapper(
                            completion_url=completion_url,
                            completionID=completionID,
                            response=response,
                            partial_log_row=log_row,
                        )
                    response = output.copy()

                    if "choices" in response:
                        response = flatten_response(response)
                elif "lamini" in func.__module__:
                    response = {
                        "id": str(uuid.uuid4()),
                        "object": "chat.completion",
                        "choices": [
                            {
                                "index": 0,
                                "finish_reason": "stop",
                                "message": {
                                    "role": "assistant",
                                    "content": output["output"],
                                },
                            }
                        ],
                        "usage": {
                            "prompt_tokens": 0,
                            "completion_tokens": 0,
                            "total_tokens": 0,
                        },
                    }
                elif "mistralai" in func.__module__:
                    if "stream" in func.__qualname__:
                        log_row["response"] = response
                        log_row["status"] = "finished"
                        return StreamingResponseWrapper(
                            completion_url=completion_url,
                            completionID=completionID,
                            response=response,
                            partial_log_row=log_row,
                        )
                    response = output.copy()

                if hasattr(response, "model_dump_json"):
                    response = response.model_dump_json()
                else:
                    response = json.dumps(response)

                log_row["status"] = "finished"
                log_row["duration"] = int(duration * 1000)
                log_row["response"] = response

                if target_service == "log10":
                    try:
                        _url = f"{completion_url}/{completionID}"
                        res = post_request(_url, log_row)
                        if res.status_code != 200:
                            logger.error(f"LOG10: failed to insert in log10: {log_row} with error {res.text}")
                    except Exception as e:
                        logging.warn(f"LOG10: failed to log: {e}. Skipping")

                elif target_service == "bigquery":
                    try:
                        log_row["id"] = str(uuid.uuid4())
                        log_row["created_at"] = datetime.now(timezone.utc).isoformat()
                        log_row["request"] = json.dumps(kwargs)

                        if func.__qualname__ == "Completion.create":
                            log_row["kind"] = "completion"
                        elif func.__qualname__ == "ChatCompletion.create":
                            log_row["kind"] = "chat"

                        log_row["orig_module"] = func.__module__
                        log_row["orig_qualname"] = func.__qualname__
                        log_row["session_id"] = session_id_var.get()

                        bigquery_client.insert_rows_json(bigquery_table, [log_row])

                    except Exception as e:
                        logging.error(f"LOG10: failed to insert in Bigquery: {log_row} with error {e}")

        return output

    return wrapper


def set_sync_log_text(USE_ASYNC=True):
    return "async" if USE_ASYNC else "sync"


def log10(module, DEBUG_=False, USE_ASYNC_=True):
    """Intercept and overload module for logging purposes
    support both openai V0 and V1, anthropic, vertexai, and mistralai

    Keyword arguments:
    module -- the module to be intercepted (e.g. openai)
    DEBUG_ -- whether to show log10 related debug statements via python logging (default False)
    USE_ASYNC_ -- whether to run in async mode (default True)

    Openai V0 example:
    Example:
        >>> # xdoctest: +SKIP
        >>> from log10.load import log10
        >>> import openai
        >>> log10(openai)
        >>> completion = openai.Completion.create(
        >>>     model="gpt-3.5-turbo-instruct",
        >>>     prompt="Once upon a time",
        >>>     max_tokens=32,
        >>> )
        >>> print(completion)

    Example:
        >>> # xdoctest: +SKIP
        >>> from log10.load import log10
        >>> import openai
        >>> log10(openai)
        >>> completion = openai.ChatCompletion.create(
        >>>     model="gpt-3.5-turbo",
        >>>     messages=[{"role": "user", "content": "Hello world"}],
        >>>     max_tokens=8,
        >>> )
        >>> print(completion)

    Example:
        >>> from log10.load import log10
        >>> from langchain.chat_models import ChatOpenAI
        >>> from langchain.schema import HumanMessage, SystemMessage
        >>> import openai
        >>> log10(openai)
        >>> llm = ChatOpenAI(
        >>>     model_name="gpt-3.5-turbo",
        >>>     temperature=0.5,
        >>> )
        >>> messages = [
        >>>     SystemMessage(content="You are a ping pong machine"),
        >>>     HumanMessage(content="Ping?")
        >>> ]
        >>> completion = llm.predict_messages(messages)
        >>> print(completion)

    Example:
        >>> from log10.load import log10
        >>> import anthropic
        >>> log10(anthropic)
        >>> from anthropic import Anthropic, HUMAN_PROMPT, AI_PROMPT
        >>> anthropic = Anthropic()
        >>> completion = anthropic.completions.create(
        >>>     model="claude-1",
        >>>     max_tokens_to_sample=32,
        >>>     prompt=f"{HUMAN_PROMPT} Hi, how are you? {AI_PROMPT}",
        >>> )
        >>> print(completion.completion)

    Example:
        >>> from log10.load import log10
        >>> import anthropic
        >>> from langchain.chat_models import ChatAnthropic
        >>> from langchain.schema import HumanMessage, SystemMessage
        >>> log10(anthropic)
        >>> llm = ChatAnthropic(model="claude-1", temperature=0.7)
        >>> messages = [
        >>>     SystemMessage(content="You are a ping pong machine"),
        >>>     HumanMessage(content="Ping?")
        >>> ]
        >>> completion = llm.predict_messages(messages)
        >>> print(completion)
    """
    global DEBUG, USE_ASYNC, sync_log_text
    DEBUG = DEBUG_ or os.environ.get("LOG10_DEBUG", False)
    logger.setLevel(logging.DEBUG if DEBUG else logging.WARNING)
    USE_ASYNC = USE_ASYNC_
    sync_log_text = set_sync_log_text(USE_ASYNC=USE_ASYNC)
    if DEBUG:
        httpx_logger = logging.getLogger("httpx")
        httpx_logger.setLevel(logging.DEBUG)

    # def intercept_nested_functions(obj):
    #     for name, attr in vars(obj).items():
    #         if callable(attr) and isinstance(attr, types.FunctionType):
    #             setattr(obj, name, intercepting_decorator(attr))
    #         elif inspect.isclass(attr):
    #             intercept_class_methods(attr)

    # def intercept_class_methods(cls):
    #     for method_name, method in vars(cls).items():
    #         if isinstance(method, classmethod):
    #             original_method = method.__func__
    #             decorated_method = intercepting_decorator(original_method)
    #             setattr(cls, method_name, classmethod(decorated_method))
    #         elif isinstance(method, (types.FunctionType, types.MethodType)):
    #             print(f"method:{method}")
    #             setattr(cls, method_name, intercepting_decorator(method))
    #         elif inspect.isclass(method):  # Handle nested classes
    #             intercept_class_methods(method)

    if getattr(module, "_log10_patched", False):
        logger.warning(f"{module.__name__} already patched. Skipping.")
        return

    if module.__name__ == "anthropic":
        from log10._httpx_utils import InitPatcher

        # Patch the AsyncAnthropic and Anthropic class
        InitPatcher(module, ["AsyncAnthropic", "Anthropic"])
    elif module.__name__ == "lamini":
        attr = module.api.utils.completion.Completion
        method = getattr(attr, "generate")
        setattr(attr, "generate", intercepting_decorator(method))
    elif module.__name__ == "mistralai" and getattr(module, "_log10_patched", False) is False:
        attr = module.client.MistralClient
        method = getattr(attr, "chat")
        setattr(attr, "chat", intercepting_decorator(method))

        method = getattr(attr, "chat_stream")
        setattr(attr, "chat_stream", intercepting_decorator(method))
    elif module.__name__ == "openai":
        openai_version = parse(version("openai"))
        global OPENAI_V1
        OPENAI_V1 = openai_version >= parse("1.0.0")

        # support for sync completions
        if OPENAI_V1:
            attr = module.resources.completions.Completions
            method = getattr(attr, "create")
            setattr(attr, "create", intercepting_decorator(method))

            attr = module.resources.chat.completions.Completions
            method = getattr(attr, "create")
            setattr(attr, "create", intercepting_decorator(method))

            # support for async completions
            from log10._httpx_utils import InitPatcher

            # Patch the AsyncOpenAI class
            InitPatcher(module, ["AsyncOpenAI"])
        else:
            attr = module.api_resources.completion.Completion
            method = getattr(attr, "create")
            setattr(attr, "create", intercepting_decorator(method))

            attr = module.api_resources.chat_completion.ChatCompletion
            mothod = getattr(attr, "create")
            setattr(attr, "create", intercepting_decorator(mothod))
    elif module.__name__ == "vertexai":
        # patch chat _send_message function
        attr = module.generative_models._generative_models.ChatSession
        method = getattr(attr, "_send_message")
        setattr(attr, "_send_message", intercepting_decorator(method))
    elif module.__name__ == "google.generativeai":
        # patch ChatSession send_message function
        attr = module.ChatSession
        method = getattr(attr, "send_message")
        setattr(attr, "send_message", intercepting_decorator(method))
    else:
        logger.warning(f"Unsupported module: {module.__name__}. Please contact us for support at support@log10.io.")

        # For future reference:
        # if callable(attr) and isinstance(attr, types.FunctionType):
        #     print(f"attr:{attr}")
        #     setattr(module, name, intercepting_decorator(attr))
        # elif inspect.isclass(attr):  # Check if attribute is a class
        #     intercept_class_methods(attr)
        # # else: # uncomment if we want to include nested function support
        # #     intercept_nested_functions(attr)

    module._log10_patched = True


if is_openai_v1():
    import openai
    from openai import OpenAI

    class OpenAI(OpenAI):
        """
        Example:
            >>> from log10.load import OpenAI
            >>> client = OpenAI(tags=["load_v1_test"])
            >>> completion = client.completions.create(model='gpt-3.5-turbo-instruct', prompt="Twice upon a time", max_tokens=32)
            >>> print(completion)

        Example:
            >>> from log10.load import OpenAI
            >>> client = OpenAI(tags=["load_v1_test"])
            >>> completion = client.chat.completions.create(
            >>>     model="gpt-3.5-turbo",
            >>>     messages=[{"role": "user", "content": "Hello world"}],
            >>> )
            >>> print(completion)
        """

        def __init__(self, *args, **kwargs):
            # check if tags is passed in
            if "tags" in kwargs:
                tags_var.set(kwargs.pop("tags"))
            super().__init__(*args, **kwargs)

            if not getattr(openai, "_log10_patched", False):
                log10(openai)
                openai._log10_patched = True


try:
    import anthropic
except ImportError:
    logger.debug("Anthropic not found. Skipping defining log10.load.Anthropic client.")
else:
    from anthropic import Anthropic, AsyncAnthropic

    class _Log10Anthropic:
        def __init__(self, *args, **kwargs):
            if "tags" in kwargs:
                tags_var.set(kwargs.pop("tags"))

            if not getattr(anthropic, "_log10_patched", False):
                log10(anthropic)
                anthropic._log10_patched = True

            super().__init__(*args, **kwargs)

    class Anthropic(_Log10Anthropic, Anthropic):
        """
        Example:
            >>> from log10.load import Anthropic
            >>> client = Anthropic(tags=["test", "load_anthropic"])
            >>> message = client.messages.create(
            ...     model="claude-3-haiku-20240307",
            ...     max_tokens=100,
            ...     temperature=0.9,
            ...     system="Respond only in Yoda-speak.",
            ...     messages=[{"role": "user", "content": "How are you today?"}],
            ... )
            >>> print(message.content[0].text)
        """

        pass

    class AsyncAnthropic(_Log10Anthropic, AsyncAnthropic):
        """
        Example:
            >>> import asyncio
            >>> from log10._httpx_utils import finalize
            >>> from log10.load import AsyncAnthropic
            >>> client = AsyncAnthropic(tags=["test", "async_anthropic"])
            >>> async def main() -> None:
            >>>     message = await client.messages.create(
            ...         model="claude-3-haiku-20240307",
            ...         max_tokens=100,
            ...         temperature=0.9,
            ...         system="Respond only in Yoda-speak.",
            ...         messages=[{"role": "user", "content": "How are you today?"}],
            ...     )
            >>>     print(message.content[0].text)
            >>>     await finalize()
            >>> asyncio.run(main())
        """

        pass



import logging

import httpx
from dotenv import load_dotenv

from log10._httpx_utils import _try_get
from log10.llm import Log10Config


load_dotenv()

logging.basicConfig(
    format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger: logging.Logger = logging.getLogger("LOG10")
logger.setLevel(logging.INFO)


class FeedbackTask:
    feedback_task_create_url = "/api/v1/feedback_task"

    def __init__(self, log10_config: Log10Config = None):
        self._log10_config = log10_config or Log10Config()
        self._http_client = httpx.Client()
        self._http_client.headers = {
            "x-log10-token": self._log10_config.token,
            "x-log10-organization-id": self._log10_config.org_id,
            "Content-Type": "application/json",
        }

    def _post_request(self, url: str, json_payload: dict) -> httpx.Response:
        json_payload["organization_id"] = self._log10_config.org_id
        try:
            res = self._http_client.post(self._log10_config.url + url, json=json_payload)
            res.raise_for_status()
            return res
        except Exception as e:
            logger.error(e)
            logger.error(e.response.json()["error"])
            raise

    def create(
        self, task_schema: dict, name: str, completion_tags_selector: list[str] = None, instruction: str = None
    ) -> httpx.Response:
        json_payload = {"json_schema": task_schema, "name": name, "completion_tags_selector": completion_tags_selector}
        if instruction:
            json_payload["instruction"] = instruction

        res = self._post_request(self.feedback_task_create_url, json_payload)
        return res

    def list(self, limit: int = 10, offset: int = 0) -> httpx.Response:
        url = f"{self._log10_config.url}/api/v1/feedback_task?organization_id={self._log10_config.org_id}&offset={offset}&limit={limit}"
        try:
            res = self._http_client.get(url=url)
            res.raise_for_status()
            return res
        except Exception as e:
            logger.error(e)
            logger.error(e.response.json()["error"])
            raise

    def get(self, id: str) -> httpx.Response:
        based_url = self._log10_config.url
        api_url = "/api/v1/feedback_task"
        get_url = f"{based_url}{api_url}/{id}?organization_id={self._log10_config.org_id}"
        res = _try_get(get_url)
        res.raise_for_status()
        if res.status_code != 200:
            raise Exception(f"Error fetching feedback task {res.json()}")
        return res



import time
from copy import deepcopy
from importlib.metadata import version
from typing import List

import openai
from packaging.version import parse

from log10.llm import LLM, ChatCompletion, Kind, Message, TextCompletion


def is_openai_v1() -> bool:
    """Return whether OpenAI API is v1 or more."""
    _version = parse(version("openai"))
    return _version.major >= 1


class OpenAI(LLM):
    def __init__(self, hparams: dict = None, log10_config=None):
        assert is_openai_v1(), "OpenAI API version must be >= 1.0.0 to use log10.OpenAI class"
        super().__init__(hparams, log10_config)
        self._client = openai.OpenAI()

    def chat(self, messages: List[Message], hparams: dict = None) -> ChatCompletion:
        """
        Example:
            >>> from log10.llm import Log10Config, Message
            >>> from log10.openai import OpenAI
            >>> llm = OpenAI({"model": "gpt-3.5-turbo"}, log10_config=Log10Config())
            >>> response = llm.chat([Message(role="user", content="Hello")])
            >>> print(response)
        """
        request = self.chat_request(messages, hparams)

        start_time = time.perf_counter()
        completion = self._client.chat.completions.create(**request)

        self.completion_id = self.log_start(request, Kind.chat)

        response = ChatCompletion(
            role=completion.choices[0].message.role,
            content=completion.choices[0].message.content,
            response=completion,
        )

        self.log_end(
            self.completion_id,
            completion.model_dump(),
            time.perf_counter() - start_time,
        )

        return response

    def chat_request(self, messages: List[Message], hparams: dict = None) -> dict:
        merged_hparams = deepcopy(self.hparams)
        if hparams:
            for hparam in hparams:
                merged_hparams[hparam] = hparams[hparam]

        return {
            "messages": [message.to_dict() for message in messages],
            **merged_hparams,
        }

    def text(self, prompt: str, hparams: dict = None) -> TextCompletion:
        """
        Example:
            >>> from log10.llm import Log10Config
            >>> from log10.openai import OpenAI
            >>> llm = OpenAI({"model": "gpt-3.5-turbo-instruct"}, log10_config=Log10Config())
            >>> response = llm.text("This is a test.")
            >>> print(response)
        """
        request = self.text_request(prompt, hparams)

        start_time = time.perf_counter()
        completion_id = self.log_start(request, Kind.text)

        completion = self._client.completions.create(**request)
        response = TextCompletion(text=completion.choices[0].text, response=completion)

        self.log_end(
            completion_id,
            completion.model_dump(),
            time.perf_counter() - start_time,
        )

        return response

    def text_request(self, prompt: str, hparams: dict = None) -> dict:
        merged_hparams = deepcopy(self.hparams)
        if hparams:
            for hparam in hparams:
                merged_hparams[hparam] = hparams[hparam]
        output = {"prompt": prompt, **merged_hparams}
        return output



import csv
import json
import os
import subprocess
import tempfile

from log10.llm import Messages
from log10.utils import fuzzy_match, parse_field


def run_metric(metric, ideal, model_output):
    ideal = parse_field(ideal)
    for ideal_candidate in ideal:
        # Ref: https://github.com/openai/evals/blob/a24f20a357ecb3cc5eec8323097aeade9585796c/evals/elsuite/utils.py
        if metric == "match":
            if model_output.startswith(ideal_candidate):
                return True
        elif metric == "includes":  # case-insensitive
            # Ref: https://github.com/openai/evals/blob/a24f20a357ecb3cc5eec8323097aeade9585796c/evals/elsuite/basic/includes.py
            if ideal_candidate.lower() in model_output.lower():
                return True
        elif metric == "fuzzy_match":
            if fuzzy_match(ideal_candidate, model_output):
                return True
    return False


def write_to_csv(file_name, row_data):
    with open(file_name, "a", newline="") as file:
        writer = csv.writer(file)
        writer.writerow(row_data)


def eval(llm, dataset, metric, out_file_path):
    csv_file_name, mapping = dataset
    with open(csv_file_name, "r") as csv_file:
        reader = csv.DictReader(csv_file)
        examples = []
        for example in reader:
            mapped_example = {}
            for key, value in mapping.items():
                mapped_example[key] = example[value]
            examples.append(mapped_example)

        # todo: each example could be launched as separate threads or separate api calls to job runners
        write_to_csv(out_file_path, ["input", "ideal", "model_completion", "metric"])
        for example in examples:
            messages = Messages.from_dict(json.loads(example["input"]))
            model_completion = llm.chat(messages)

            example_metric = run_metric(metric, example["ideal"], model_completion.content)
            write_to_csv(
                out_file_path,
                [example["input"], example["ideal"], model_completion, example_metric],
            )
            print(f"\n\nIdeal:{example['ideal']}\nModel output:{model_completion.content}")
            print(f"Correct:{example_metric}")
    return


def compile(code):
    with tempfile.NamedTemporaryFile(suffix=".c", delete=False) as temp:
        temp.write(code.encode())
        temp.close()
        process = subprocess.run(["gcc", temp.name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        os.remove(temp.name)  # remove the temp file
        if process.returncode == 0:
            return True
        else:
            return False, process.stderr.decode()



import json
from pathlib import Path

import click
import rich
from rich.console import Console
from rich.table import Table
from tqdm import tqdm

from log10.feedback.feedback import Feedback, _get_feedback_list, _get_feedback_list_graphql


@click.command()
@click.option("--task_id", prompt="Enter task id", help="Task ID")
@click.option("--values", prompt="Enter task values", help="Feedback in JSON format")
@click.option(
    "--completion_tags_selector",
    prompt="Enter completion tags selector",
    help="Completion tags selector",
)
@click.option("--comment", help="Comment", default="")
def create_feedback(task_id, values, completion_tags_selector, comment):
    """
    Add feedback to a group of completions associated with a task
    """
    click.echo("Creating feedback")
    tags = completion_tags_selector.split(",")
    values = json.loads(values)
    feedback = Feedback().create(task_id=task_id, values=values, completion_tags_selector=tags, comment=comment)
    click.echo(feedback.json())


@click.command()
@click.option(
    "--offset", default=0, type=int, help="The starting index from which to begin the feedback fetch. Defaults to 0."
)
@click.option(
    "--limit", default=25, type=int, help="The maximum number of feedback items to retrieve. Defaults to 25."
)
@click.option(
    "--task_id",
    default="",
    type=str,
    help="The specific Task ID to filter feedback. If not provided, feedback for all tasks will be fetched.",
)
@click.option(
    "--filter",
    default="",
    type=str,
    help="The filter applied to the feedback. If not provided, feedback will not be filtered. e.g. `log10 feedback list --filter 'Coverage <= 5'`.",
)
def list_feedback(offset, limit, task_id, filter):
    """
    List feedback based on the provided criteria. This command allows fetching feedback for a specific task or across all tasks,
    with control over the starting point and the number of items to retrieve.
    """
    feedback_data = (
        _get_feedback_list_graphql(task_id, filter, page=1, limit=limit)
        if filter
        else _get_feedback_list(offset, limit, task_id)
    )
    data_for_table = []
    for feedback in feedback_data:
        data_for_table.append(
            {
                "id": feedback["id"],
                "task_name": feedback["task_name"],
                "feedback": json.dumps(feedback["json_values"], ensure_ascii=False),
                "matched_completion_ids": ",".join(feedback["matched_completion_ids"]),
            }
        )
    table = Table(title="Feedback")
    table.add_column("ID")
    table.add_column("Task Name")
    table.add_column("Feedback")
    table.add_column("Completion ID")

    for item in data_for_table:
        table.add_row(item["id"], item["task_name"], item["feedback"], item["matched_completion_ids"])
    console = Console()
    console.print(table)
    console.print(f"Total feedback: {len(feedback_data)}")


@click.command()
@click.option("--id", required=True, help="Get feedback by ID")
def get_feedback(id):
    """
    Get feedback based on provided ID.
    """
    try:
        res = Feedback().get(id)
    except Exception as e:
        click.echo(f"Error fetching feedback {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            click.echo(e.response.json()["error"])
        return
    console = Console()
    feedback = json.dumps(res.json(), indent=4)
    console.print_json(feedback)


@click.command()
@click.option(
    "--offset", default=0, type=int, help="The starting index from which to begin the feedback fetch. Defaults to 0."
)
@click.option(
    "--limit", default=25, type=int, help="The maximum number of feedback items to retrieve. Defaults to 25."
)
@click.option(
    "--task_id",
    default="",
    type=str,
    help="The specific Task ID to filter feedback. If not provided, feedback for all tasks will be fetched.",
)
@click.option(
    "--filter",
    default="",
    type=str,
    help="The filter applied to the feedback. If not provided, feedback will not be filtered. e.g. `log10 feedback list --filter 'Coverage <= 5'`.",
)
@click.option(
    "--file",
    "-f",
    type=click.Path(dir_okay=False),
    required=False,
    help="Path to the file where the feedback will be saved. The feedback data is saved in JSON Lines (jsonl) format. If not specified, feedback will be printed to stdout.",
)
def download_feedback(offset, limit, task_id, filter, file):
    """
    Download feedback based on the provided criteria. This command allows fetching feedback for a specific task or across all tasks,
    with control over the starting point and the number of items to retrieve.
    """
    if file:
        file_path = Path(file)
        if file_path.exists():
            rich.print(f"Warning: The file {file_path} already exists and will be overwritten.")

        ext_filename = file_path.suffix.lower()
        if ext_filename != ".jsonl":
            raise click.UsageError(f"Only .jsonl extension is supported for the output file. Got: {ext_filename}")

    feedback_data = (
        _get_feedback_list_graphql(task_id, filter, page=1, limit=limit)
        if filter
        else _get_feedback_list(offset, limit, task_id)
    )

    console = Console()
    if not file:
        for feedback in feedback_data:
            console.print_json(json.dumps(feedback, indent=4))
        return

    with open(file, "w") as f:
        console.print(f"Saving feedback to {file}")
        for feedback in tqdm(feedback_data):
            f.write(json.dumps(feedback) + "\n")



import logging
from typing import List, Tuple

from dotenv import load_dotenv

from log10.llm import LLM, Message


load_dotenv()

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(levelname)s - LOG10 - CAMEL - %(message)s",
)

# Repeat word termination conditions
# https://github.com/lightaime/camel/blob/master/examples/ai_society/role_playing_multiprocess.py#L63
repeat_word_threshold = 4
repeat_word_list = ["goodbye", "good bye", "thank", "bye", "welcome", "language model"]


def camel_agent(
    user_role: str,
    assistant_role: str,
    task_prompt: str,
    max_turns: int,
    user_prompt: str = None,
    assistant_prompt: str = None,
    summary_model: str = None,
    llm: LLM = None,
) -> Tuple[List[Message], List[Message]]:
    generator = camel_agent_generator(
        user_role,
        assistant_role,
        task_prompt,
        max_turns,
        user_prompt,
        assistant_prompt,
        summary_model,
        llm,
    )
    *_, last = generator
    return last


def camel_agent_generator(
    user_role: str,
    assistant_role: str,
    task_prompt: str,
    max_turns: int,
    user_prompt: str,
    assistant_prompt: str,
    summary_model: str,
    llm: LLM,
):
    try:
        assistant_inception_prompt = f"""Never forget you are a {assistant_role} and I am a {user_role}. Never flip roles! Never instruct me!
We share a common interest in collaborating to successfully complete a task.
You must help me to complete the task.
Here is the task: {task_prompt}. Never forget our task!
I must instruct you based on your expertise and my needs to complete the task.

I must give you one instruction at a time.
You must write a specific solution that appropriately completes the requested instruction.
You must decline my instruction honestly if you cannot perform the instruction due to physical, moral, legal reasons or your capability and explain the reasons.
Do not add anything else other than your solution to my instruction.
You are never supposed to ask me any questions you only answer questions.
You are never supposed to reply with a flake solution. Explain your solutions.
Your solution must be declarative sentences and simple present tense.
Unless I say the task is completed, you should always start with:

Solution: <YOUR_SOLUTION>

<YOUR_SOLUTION> should be specific and provide preferable implementations and examples for task-solving.
Always end <YOUR_SOLUTION> with: Next request."""

        if assistant_prompt is not None:
            assistant_inception_prompt = assistant_prompt

        user_inception_prompt = f"""Never forget you are a {user_role} and I am a {assistant_role}. Never flip roles! You will always instruct me.
We share a common interest in collaborating to successfully complete a task.
I must help you to complete the task.
Here is the task: {task_prompt}. Never forget our task!
You must instruct me based on my expertise and your needs to complete the task ONLY in the following two ways:

1. Instruct with a necessary input:
Instruction: <YOUR_INSTRUCTION>
Input: <YOUR_INPUT>

2. Instruct without any input:
Instruction: <YOUR_INSTRUCTION>
Input: None

The "Instruction" describes a task or question. The paired "Input" provides further context or information for the requested "Instruction".

You must give me one instruction at a time.
I must write a response that appropriately completes the requested instruction.
I must decline your instruction honestly if I cannot perform the instruction due to physical, moral, legal reasons or my capability and explain the reasons.
You should instruct me not ask me questions.
Now you must start to instruct me using the two ways described above.
Do not add anything else other than your instruction and the optional corresponding input!
Keep giving me instructions and necessary inputs until you think the task is completed.
When the task is completed, you must only reply with a single word <CAMEL_TASK_DONE>.
Never say <CAMEL_TASK_DONE> unless my responses have solved your task."""

        if user_prompt is not None:
            user_inception_prompt = user_prompt

        assistant_prompt = f"ASSISTANT PROMPT: \n {assistant_inception_prompt}"
        user_prompt = f"USER PROMPT: \n {user_inception_prompt}"

        assistant_messages = [
            Message(role="system", content=assistant_prompt),
            Message(role="user", content=assistant_prompt),
        ]
        user_messages = [
            Message(role="system", content=user_prompt),
            Message(
                role="user",
                content=user_prompt
                + " Now start to give me instructions one by one. Only reply with Instruction and Input.",
            ),
        ]
        repeat_word_counter = 0
        repeat_word_threshold_exceeded = False

        for i in range(max_turns):
            repeated_word_current_turn = False
            #
            # User turn
            #
            user_message = llm.chat(user_messages)
            user_messages.append(user_message)
            assistant_messages.append(Message(role="user", content=user_message.content))
            logging.info(f"User turn {i}: {user_message}")

            #
            # Assistant turn
            #
            assistant_message = llm.chat(assistant_messages)
            assistant_messages.append(assistant_message)
            user_messages.append(Message(role="user", content=assistant_message.content))
            logging.info(f"Assistant turn {i}: {assistant_message}")

            yield (user_messages, assistant_messages)

            #
            # Termination conditions.
            #
            for repeat_word in repeat_word_list:
                if repeat_word in assistant_message.content.lower() or repeat_word in user_message.content.lower():
                    repeat_word_counter += 1
                    repeated_word_current_turn = True
                    logging.info(f"Repeat word counter = {repeat_word_counter}")
                    if repeat_word_counter == repeat_word_threshold:
                        repeat_word_threshold_exceeded = True
                    break

            if not repeated_word_current_turn:
                repeat_word_counter = 0

            if (
                ("CAMEL_TASK_DONE" in user_messages[-2].content)
                or (i == max_turns - 1)
                or repeat_word_threshold_exceeded
            ):
                #
                # Summary turn
                #
                summary_context = "\n".join(
                    [
                        f"{turn.role.capitalize()} ({user_role if turn.role == 'user' else assistant_role}): {turn.content}"
                        for turn in assistant_messages
                        if turn.role in ["assistant", "user"]
                    ]
                )

                logging.info(f"summary context: {summary_context}")
                summary_system_prompt = """You are an experienced solution extracting agent.
Your task is to extract full and complete solutions by looking at the conversation between a user and an assistant with particular specializations.
You should present me with a final and detailed solution purely based on the conversation.
You should present the solution as if its yours.
Use present tense and as if you are the one presenting the solution.
You should not miss any necessary details or examples.
Keep all provided explanations and codes provided throughout the conversation.
Remember your task is not to summarize rather to extract the full solution."""

                summary_closing_prompt = f"""\n\nAs a reminder, the above context is to help you provide a complete and concise solution to the following task: {task_prompt}
Do not attempt to describe the solution, but try to answer in a way such that your answer will directly solve the task - not just describe the steps required to solve the task.
Only use the provided context above and no other sources.
Even if I told you that the task is completed in the context above you should still reply with a complete solution. Never tell me the task is completed or ask for the next request, but instead replay the final solution back to me."""
                summary_prompt = f"Task:{task_prompt}\n" + summary_context + summary_closing_prompt
                summary_messages = [
                    Message(role="system", content=summary_system_prompt),
                    Message(
                        role="user",
                        content="Here is the conversation: " + summary_prompt,
                    ),
                ]

                hparams = {"model": summary_model}
                message = llm.chat(summary_messages, hparams)
                assistant_messages.append(message)
                user_messages.append(Message(role="user", content=message.content))
                logging.info(message.content)

                # End of conversation
                yield (user_messages, assistant_messages)
                break

    except Exception as e:
        logging.error("Error in CAMEL agent: ", e)



import json
import time
from pathlib import Path

import click
import pandas as pd
import rich
from rich.console import Console
from rich.table import Table

from log10._httpx_utils import _get_time_diff, _try_get
from log10.cli_utils import generate_markdown_report, generate_results_table
from log10.completions.completions import (
    Completions,
    _check_model_support,
    _compare,
    _get_completion,
    _get_completions_url,
)
from log10.llm import Log10Config
from log10.prompt_analyzer import PromptAnalyzer, convert_suggestion_to_markdown, display_prompt_analyzer_suggestions


_log10_config = Log10Config()


def _render_completions_table(completions_data):
    data_for_table = []
    for completion in completions_data:
        prompt, response = "", ""
        if completion.get("kind") == "completion":
            prompt = completion.get("request", {}).get("prompt", "")
            response_choices = completion.get("response", {}).get("choices", [])
            if response_choices:
                response = response_choices[0].get("text", "")
        elif completion.get("kind") == "chat":
            request_messages = completion.get("request", {}).get("messages", [])
            prompt = request_messages[0].get("content", "") if request_messages else ""

            response_choices = completion.get("response", {}).get("choices", [])
            if response_choices:
                # Handle 'message' and 'function_call' within the first choice safely
                first_choice = response_choices[0]
                if "message" in first_choice:
                    message = first_choice["message"]
                    response = (
                        message.get("content")
                        or (
                            message.get("tool_calls")[-1].get("function", {}).get("arguments", "")
                            if message.get("tool_calls")
                            else ""
                        )
                        or ""
                    )
                elif "function_call" in first_choice:
                    response = json.dumps(first_choice.get("function_call", {}))
        else:
            rich.print(f"Unknown completion kind: {completion['kind']} for id: {completion['id']}")

        data_for_table.append(
            {
                "id": completion["id"],
                "status": "success" if completion["status"] == "finished" else completion["status"],
                "created_at": _get_time_diff(completion["created_at"]),
                "prompt": prompt,
                "completion": response,
                "tags": [t["name"] for t in completion["tagResolved"]],
            }
        )
    # render data_for_table with rich table
    table = Table(show_header=True, header_style="bold magenta")

    table.add_column("ID", style="dim")
    table.add_column("Status")
    table.add_column("Created At")
    table.add_column("Prompt", overflow="fold")
    table.add_column("Completion", overflow="fold")
    table.add_column("Tags", justify="right")

    max_len = 40
    for item in data_for_table:
        tags = ", ".join(item["tags"]) if item["tags"] else ""
        if isinstance(item["prompt"], list):
            item["prompt"] = " ".join(item["prompt"])
        short_prompt = item["prompt"][:max_len] + "..." if len(item["prompt"]) > max_len else item["prompt"]
        completion = item.get("completion", "")
        short_completion = completion[:max_len] + "..." if len(completion) > max_len else completion
        table.add_row(item["id"], item["status"], item["created_at"], short_prompt, short_completion, tags)

    console = Console()
    console.print(table)


def _render_comparison_table(model_response_raw_data):
    rich.print(f"completion_id: {model_response_raw_data['completion_id']}")
    rich.print(f"tags: {model_response_raw_data['tags']}")
    rich.print("original_request:")
    rich.print_json(json.dumps(model_response_raw_data["original_request"], indent=4))

    table = rich.table.Table(show_header=True, header_style="bold magenta", box=rich.box.ROUNDED, show_lines=True)
    table.add_column("Model")
    table.add_column("Content")
    table.add_column("Total Token Usage (Input/Output)")
    table.add_column("Duration (ms)")

    for model, data in model_response_raw_data.items():
        # only display model data
        if model not in ["completion_id", "original_request", "tags"]:
            usage = data["usage"]
            formatted_usage = f"{usage['total_tokens']} ({usage['prompt_tokens']}/{usage['completion_tokens']})"
            table.add_row(model, data["content"], formatted_usage, str(data["duration"]))
    rich.print(table)


def _create_dataframe_from_comparison_data(model_response_raw_data):
    completion_id = model_response_raw_data["completion_id"]
    tags = model_response_raw_data["tags"]
    original_request = model_response_raw_data["original_request"]
    rows = []
    for model, model_data in model_response_raw_data.items():
        # only display model data
        if model not in ["completion_id", "original_request", "tags"]:
            content = model_data["content"]
            usage = model_data["usage"]
            prompt_tokens = usage["prompt_tokens"]
            completion_tokens = usage["completion_tokens"]
            total_tokens = usage["total_tokens"]
            duration = model_data["duration"]
            prompt_messages = json.dumps(original_request["messages"])
            rows.append(
                [
                    completion_id,
                    tags,
                    prompt_messages,
                    model,
                    content,
                    prompt_tokens,
                    completion_tokens,
                    total_tokens,
                    duration,
                ]
            )

    df = pd.DataFrame(
        rows,
        columns=[
            "completion_id",
            "tags",
            "prompt_messages",
            "model",
            "content",
            "prompt_tokens",
            "completion_tokens",
            "total_tokens",
            "duration_ms",
        ],
    )

    return df


@click.command()
@click.option("--limit", default=25, help="Specify the maximum number of completions to retrieve.")
@click.option("--offset", default=0, help="Set the starting point (offset) from where to begin fetching completions.")
@click.option(
    "--timeout", default=10, help="Set the maximum time (in seconds) allowed for the HTTP request to complete."
)
@click.option("--tags", default="", help="Filter completions by specific tags. Separate multiple tags with commas.")
@click.option(
    "--from",
    "from_date",
    type=click.DateTime(),
    help="Define the start date for fetching completions (inclusive). Use the format: YYYY-MM-DD.",
)
@click.option(
    "--to",
    "to_date",
    type=click.DateTime(),
    help="Set the end date for fetching completions (inclusive). Use the format: YYYY-MM-DD.",
)
def list_completions(limit, offset, timeout, tags, from_date, to_date):
    """
    List completions
    """
    base_url = _log10_config.url
    org_id = _log10_config.org_id

    url = _get_completions_url(limit, offset, tags, from_date, to_date, base_url, org_id)
    # Fetch completions
    res = _try_get(url, timeout)

    completions = res.json()
    completions = completions["data"]

    _render_completions_table(completions)


@click.command()
@click.option("--id", prompt="Enter completion id", help="Completion ID")
def get_completion(id):
    """
    Get a completion by id
    """
    res = _get_completion(id)
    rich.print_json(json.dumps(res.json()["data"], indent=4))


@click.command()
@click.option("--limit", default=50, help="Specify the maximum number of completions to retrieve.")
@click.option("--offset", default=0, help="Set the starting point (offset) from where to begin fetching completions.")
@click.option(
    "--timeout", default=10, help="Set the maximum time (in seconds) allowed for the HTTP request to complete."
)
@click.option("--tags", default="", help="Filter completions by specific tags. Separate multiple tags with commas.")
@click.option(
    "--from",
    "from_date",
    type=click.DateTime(),
    help="Define the start date for fetching completions (inclusive). Use the format: YYYY-MM-DD.",
)
@click.option(
    "--to",
    "to_date",
    type=click.DateTime(),
    help="Set the end date for fetching completions (inclusive). Use the format: YYYY-MM-DD.",
)
@click.option(
    "--file",
    "-f",
    type=click.Path(dir_okay=False),
    default="completions.jsonl",
    help="Specify the filename and path for the output file. Only .jsonl extension is supported.",
)
def download_completions(limit, offset, timeout, tags, from_date, to_date, file):
    """
    Download completions to a jsonl file
    """
    input_offset = int(offset)
    input_limit = int(limit)
    fetched_total = 0
    batch_size = 10

    if file:
        path = Path(file)
        if path.exists():
            rich.print(f'Warning: The file "{file}" already exists and will be overwritten.')

        ext_name = path.suffix.lower()
        if ext_name not in [".jsonl"]:
            raise click.UsageError(f"Only .jsonl extension is supported for the output file. Got: {ext_name}")

    console = Console()
    track_limit = input_limit if input_limit < batch_size else batch_size
    track_offset = input_offset
    try:
        with console.status("[bold green]Downloading completions...", spinner="bouncingBar") as status:
            with open(file, "w") as output_file:
                start_time = time.time()
                while True and track_limit > 0:
                    new_data = Completions()._get_completions(
                        offset=track_offset,
                        limit=track_limit,
                        timeout=timeout,
                        tag_names=tags,
                        from_date=from_date,
                        to_date=to_date,
                    )

                    new_data_size = len(new_data)
                    fetched_total += new_data_size

                    for completion in new_data:
                        res = _get_completion(completion["id"])
                        if res.status_code != 200:
                            rich.print(f"Error fetching completion {completion['id']}")
                            continue
                        if not res.json().get("data", {}):
                            rich.print(f"Completion {completion['id']} is empty")
                            continue
                        full_completion_data = res.json()["data"]
                        output_file.write(json.dumps(full_completion_data) + "\n")

                    elapsed_time = time.time() - start_time
                    rate = fetched_total / elapsed_time if elapsed_time > 0 else 0
                    status.update(
                        f"[bold green]Downloading completions...\n"
                        f"📥 Downloaded {fetched_total} | "
                        f"⏱️ {elapsed_time:.1f}s | "
                        f"⚡ {rate:.1f}/s"
                    )

                    if new_data_size == 0 or new_data_size < track_limit:
                        break

                    track_offset += new_data_size
                    track_limit = (
                        input_limit - fetched_total if input_limit - fetched_total < batch_size else batch_size
                    )
    except Exception as e:
        rich.print(f"Error fetching completions {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            rich.print(e.response.json()["error"])
        return

    rich.print(
        f"[bold green]📥 Downloaded {fetched_total} | "
        f"⏱️ {elapsed_time:.1f}s. | "
        f"⚡ {rate:.2f}/s\n"
        f"💾 Saved to {file}"
    )


@click.command()
@click.option(
    "--ids",
    default="",
    help="Log10 completion IDs. Provide a comma-separated list of completion IDs or a "
    "path to a JSON file containing the list of IDs.",
)
@click.option("--tags", default="", help="Filter completions by specific tags. Separate multiple tags with commas.")
@click.option("--limit", help="Specify the maximum number of completions to retrieve filtered by tags.")
@click.option(
    "--offset", help="Set the starting point (offset) from where to begin fetching completions filtered by tags."
)
@click.option("--models", default="", help="Comma separated list of models to compare")
@click.option("--temperature", default=0.2, help="Temperature")
@click.option("--max_tokens", default=512, help="Max tokens")
@click.option("--top_p", default=1.0, help="Top p")
@click.option("--analyze_prompt", is_flag=True, help="Run prompt analyzer on the messages.")
@click.option(
    "--file",
    "-f",
    type=click.Path(dir_okay=False),
    help="Specify the filename to save the results. Specify the output file using `.md` for a markdown report, "
    "`.csv` for comma-separated values, or `.jsonl` for JSON Lines format. Only .md, .csv, and .jsonl extensions "
    "are supported.",
)
def benchmark_models(
    ids,
    tags,
    limit,
    offset,
    models,
    temperature,
    max_tokens,
    top_p,
    file,
    analyze_prompt,
):
    """
    Compare completions using different models and generate report
    """
    if ids and tags:
        raise click.UsageError("--ids and --tags cannot be set together.")
    if (limit or offset) and not tags:
        raise click.UsageError("--limit and --offset can only be used with --tags.")
    if tags:
        if not limit:
            limit = 5
        if not offset:
            offset = 0

    if not models:
        raise click.UsageError("--models must be set to compare.")
    else:
        for model in [m for m in models.split(",") if m]:
            if not _check_model_support(model):
                raise click.UsageError(f"Model {model} is not supported.")

    if file:
        path = Path(file)
        ext_name = path.suffix.lower()
        if ext_name not in [".md", ".csv", ".jsonl"]:
            raise click.UsageError(
                f"Only .md, .csv, and .jsonl extensions are supported for the output file. Got: {ext_name}"
            )

    # get completions ids
    completion_ids = []
    if ids:
        ids_path = Path(ids)
        if ids_path.is_file():
            if ids_path.suffix.lower() != ".json":
                raise click.UsageError(
                    f"Only json format is supported for the IDs file. Got {ids_path.suffix.lower()}"
                )
            with open(ids_path, "r") as f:
                completion_ids = json.load(f)
        else:
            completion_ids = [id for id in ids.split(",") if id]
    elif tags:
        base_url = _log10_config.url
        org_id = _log10_config.org_id
        url = _get_completions_url(limit, offset, tags, None, None, base_url, org_id)
        res = _try_get(url)
        completions = res.json()["data"]
        completion_ids = [completion["id"] for completion in completions]
        if not completion_ids:
            SystemExit(f"No completions found for tags: {tags}")

    compare_models = [m for m in models.split(",") if m]

    data = []
    skipped_completion_ids = []
    for id in completion_ids:
        rich.print(f"Processing completion {id}")
        # get message from id
        completion_data = _get_completion(id).json()["data"]
        tags = [t["name"] for t in completion_data["tagResolved"]]

        # skip completion if status is not finished or kind is not chat
        if completion_data["status"] != "finished" or completion_data["kind"] != "chat":
            rich.print(f"Skip completion {id}. Status is not finished or kind is not chat.")
            skipped_completion_ids.append(id)
            continue

        original_model_request = completion_data["request"]
        original_model_response = completion_data["response"]
        original_model = original_model_response["model"]
        benchmark_data = {
            "completion_id": id,
            "original_request": original_model_request,
            f"{original_model} (original model)": {
                "content": original_model_response["choices"][0]["message"]["content"],
                "usage": original_model_response["usage"],
                "duration": completion_data["duration"],
            },
            "tags": tags,
        }
        messages = original_model_request["messages"]
        compare_models_data = _compare(compare_models, messages, temperature, max_tokens, top_p)
        benchmark_data.update(compare_models_data)
        data.append(benchmark_data)

    prompt_analysis_data = {}
    if analyze_prompt:
        rich.print("Analyzing prompts")
        for item in data:
            completion_id = item["completion_id"]
            prompt_messages = item["original_request"]["messages"]
            all_messages = "\n\n".join([m["content"] for m in prompt_messages])
            analyzer = PromptAnalyzer()
            suggestions = analyzer.analyze(all_messages)
            prompt_analysis_data[completion_id] = suggestions

    # create an empty dataframe
    all_df = pd.DataFrame(
        columns=[
            "completion_id",
            "tags",
            "prompt_messages",
            "model",
            "content",
            "prompt_tokens",
            "completion_tokens",
            "total_tokens",
            "duration_ms",
        ]
    )

    #
    # Display or save the results
    #
    if not file:
        # display in terminal using rich
        for ret in data:
            _render_comparison_table(ret)
            if analyze_prompt:
                completion_id = ret["completion_id"]
                suggestions = prompt_analysis_data[completion_id]
                rich.print(f"Prompt Analysis for completion_id: {completion_id}")
                display_prompt_analyzer_suggestions(suggestions)
    else:
        # generate markdown report and save to file
        for ret in data:
            df = _create_dataframe_from_comparison_data(ret)
            all_df = pd.concat([all_df, df])

        if ext_name == ".csv":
            # save the dataframe to a csv file
            all_df.to_csv(path, index=False)
            rich.print(f"Dataframe saved to csv file: {path}")
        elif ext_name == ".jsonl":
            # save the dataframe to a jsonl file
            all_df.to_json(path, orient="records", lines=True)
            rich.print(f"Dataframe saved to jsonl file: {path}")
        elif ext_name == ".md":
            # generate markdown report
            pivot_df = all_df.pivot(index="completion_id", columns="model", values="content")
            pivot_df["prompt_messages"] = all_df.groupby("completion_id")["prompt_messages"].first()
            # Reorder the columns
            cols = pivot_df.columns.tolist()
            cols = [cols[-1]] + cols[:-1]
            pivot_df = pivot_df[cols]

            pivot_table = generate_results_table(pivot_df, section_name="model comparison")
            all_results_table = generate_results_table(all_df, section_name="All Results")

            prompt_analysis_markdown = ""
            if analyze_prompt:
                prompt_analysis_markdown = "## Prompt Analysis\n\n"
                for completion_id, suggestions in prompt_analysis_data.items():
                    prompt_messages = all_df[all_df["completion_id"] == completion_id]["prompt_messages"].values[0]
                    prompt_analysis_markdown += (
                        f"### Prompt Analysis for completion_id: {completion_id}\n\n{prompt_messages}\n\n"
                    )
                    prompt_analysis_markdown += convert_suggestion_to_markdown(suggestions)

            # generate the list of skipped completions ids
            skipped_completion_markdown = ""
            if skipped_completion_ids:
                skipped_completion_ids_str = ", ".join(skipped_completion_ids)
                skipped_completion_markdown += "## Skipped Completion IDs\n\n"
                skipped_completion_markdown += f"Skipped completions: {skipped_completion_ids_str}\n\n"

            generate_markdown_report(
                path, [pivot_table, prompt_analysis_markdown, all_results_table, skipped_completion_markdown]
            )
            rich.print(f"Report saved to markdown file: {path}")



import json

import click
import rich
from rich.console import Console

from log10.feedback.autofeedback import AutoFeedbackICL, get_autofeedback


@click.command()
@click.option("--task_id", help="Feedback task ID")
@click.option("--content", help="Completion content")
@click.option("--file", "-f", help="File containing completion content")
@click.option("--completion_id", help="Completion ID")
@click.option("--num_samples", default=5, help="Number of samples to use for few-shot learning")
def auto_feedback_icl(task_id: str, content: str, file: str, completion_id: str, num_samples: int):
    """
    Generate feedback with existing human feedback based on in context learning
    """
    options_count = sum([1 for option in [content, file, completion_id] if option])
    if options_count > 1:
        click.echo("Only one of --content, --file, or --completion_id should be provided.")
        return

    console = Console()
    auto_feedback_icl = AutoFeedbackICL(task_id, num_samples=num_samples)
    if completion_id:
        results = auto_feedback_icl.predict(completion_id=completion_id)
        console.print_json(results)
        return

    if file:
        with open(file, "r") as f:
            content = f.read()
    results = auto_feedback_icl.predict(text=content)
    console.print_json(results)


@click.command()
@click.option("--completion-id", required=True, help="Completion ID")
def get_autofeedback_cli(completion_id: str):
    """
    Get an auto feedback by completion id
    """
    res = get_autofeedback(completion_id)
    if res:
        rich.print_json(json.dumps(res["data"], indent=4))



from langchain.document_loaders import WebBaseLoader

from log10.llm import Message


# Browser agent / tool
def browser(URL: str) -> str:
    """useful when you need to scrape a website.
    Input is the URL of the website to be scraped.
    Output is the text of the website
    """
    # Example code if we want to follow links
    # explored_sites = set()
    # queue = [URL]

    # def parse_site(base_url):
    #   if base_url in explored_sites:
    #     return

    #   explored_sites.add(base_url)

    #   if debug:
    #     print("Scraping: ", base_url)

    #   page = requests.get(URL)
    #   soup = BeautifulSoup(page.content, "html.parser")
    #   for a_href in soup.find_all("a", href=True):
    #       link = a_href["href"]
    #       if link[0] == "/":
    #         link = URL + link

    #       if link.startswith(URL):
    #         to_explore = link.split("#")[0]
    #         queue.append(to_explore)

    # while len(queue) != 0:
    #   url = queue.pop()
    #   parse_site(url)

    # return list(explored_sites)

    loader = WebBaseLoader(URL)
    data = loader.load()
    data[0].page_content = " ".join(data[0].page_content.split())

    # hack: crop to 5000 char to fit within context length
    # Will need better strategy eventually
    #   maxlen = 5000 if len(data[0].page_content) > 5000 else len(data[0].page_content)
    #   return data[0].page_content[:maxlen]
    return data[0].page_content


def code_extractor(full_response, language, extraction_model, llm):
    """useful when you need to extract just the code from a detailed LLM response"""
    messages = [
        Message(
            role="system",
            content=f"Extract just the {language} code from the following snippet. Do not include ``` or markdown syntax. Format your reply so I can directly copy your entire response, save it as a file and compile and run the code.",
        ),
        Message(role="user", content="Here's the snippet:\n" + full_response),
    ]
    completion = llm.chat(messages, {"model": extraction_model, "temperature": 0.2})
    return completion.content



import logging

import httpx

from log10._httpx_utils import _try_get, _try_post_graphql_request
from log10.llm import Log10Config
from log10.utils import safe_get


logging.basicConfig(
    format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger: logging.Logger = logging.getLogger("LOG10")
logger.setLevel(logging.INFO)


class Feedback:
    feedback_create_url = "/api/v1/feedback"

    def __init__(self, log10_config: Log10Config = None):
        self._log10_config = log10_config or Log10Config()
        self._http_client = httpx.Client()
        self._http_client.headers = {
            "x-log10-token": self._log10_config.token,
            "x-log10-organization-id": self._log10_config.org_id,
            "Content-Type": "application/json",
        }

    def _post_request(self, url: str, json_payload: dict) -> httpx.Response:
        json_payload["organization_id"] = self._log10_config.org_id
        try:
            res = self._http_client.post(self._log10_config.url + url, json=json_payload)
            res.raise_for_status()
            return res
        except Exception as e:
            logger.error(e)
            if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
                logger.error(e.response.json()["error"])
            raise

    def create(
        self,
        task_id: str,
        values: dict,
        completion_tags_selector: list[str],
        comment: str = None,
    ) -> httpx.Response:
        json_payload = {
            "task_id": task_id,
            "json_values": values,
            "completion_tags_selector": completion_tags_selector,
            "comment": comment,
        }
        res = self._post_request(self.feedback_create_url, json_payload)
        return res

    def list(self, offset: int = 0, limit: int = 50, task_id: str = "", filter: str = "") -> httpx.Response:
        if filter:
            return self.list_v2(page=round(offset / limit) + 1, limit=limit, task_id=task_id, filter=filter)

        base_url = self._log10_config.url
        api_url = "/api/v1/feedback"
        url = f"{base_url}{api_url}?organization_id={self._log10_config.org_id}&offset={offset}&limit={limit}&task_id={task_id}"
        logger.debug(f"Fetching feedback from url: {url}")

        # GET feedback
        try:
            res = self._http_client.get(url=url)
            res.raise_for_status()
            return res
        except Exception as e:
            logger.error(e)
            if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
                logger.error(e.response.json()["error"])
            raise

    def list_v2(
        self, page: int = 1, limit: int = 50, task_id: str | None = None, filter: str | None = None
    ) -> httpx.Response:
        query = """
        query OrganizationFeedback($id: String!, $filter: String, $taskId: String, $page: Int, $limit: Int) {
            organization(id: $id) {
                id
                feedbackV2(filter: $filter, taskId: $taskId, page: $page, limit: $limit) {
                    pageInfo{
                        totalCount
                        currentPage
                    }
                    nodes {
                        id
                        jsonValues
                        task {
                            id
                            name
                        }
                        completions {
                            id
                        }
                    }
                }
            }
        }
        """

        variables = {
            "id": self._log10_config.org_id,
            "taskId": task_id,
            "filter": filter,
            "page": page,
            "limit": limit,
        }
        logger.debug(f"Fetching feedback with variables: {variables}")

        response = _try_post_graphql_request(query, variables)

        if response is None:
            logger.error("Failed to get feedback")
            return None

        if response.status_code == 200:
            return response.json()
        else:
            response.raise_for_status()

    def get(self, id: str) -> httpx.Response:
        base_url = self._log10_config.url
        api_url = "/api/v1/feedback"
        get_url = f"{base_url}{api_url}/{id}?organization_id={self._log10_config.org_id}"
        res = _try_get(get_url)
        res.raise_for_status()
        if res.status_code != 200:
            raise Exception(f"Error fetching feedback: {res.json()}")
        return res


def _get_feedback_list(offset, limit, task_id):
    total_fetched = 0
    feedback_data = []
    total_feedback = 0
    if limit:
        limit = int(limit)

    try:
        while True:
            fetch_limit = limit - total_fetched if limit else 50
            res = Feedback().list(offset=offset, limit=fetch_limit, task_id=task_id)
            new_data = res.json().get("data", [])
            if total_feedback == 0:
                total_feedback = res.json().get("total", 0)
            if not limit:
                limit = total_feedback
            feedback_data.extend(new_data)

            current_fetched = len(new_data)
            total_fetched += current_fetched
            offset += current_fetched
            if total_fetched >= limit or total_fetched >= total_feedback:
                break
    except Exception as e:
        logger.error(f"Error fetching feedback {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            logger.error(e.response.json()["error"])
        return []

    return feedback_data


def _format_graphql_node(node):
    return {
        "id": node["id"],
        "json_values": node["jsonValues"],
        "task_id": node["task"]["id"],
        "task_name": node["task"]["name"],
        "matched_completion_ids": [c["id"] for c in node["completions"]],
    }


def _get_feedback_list_graphql(task_id, filter, page=1, limit=50):
    feedback_data = []
    current_page = page
    limit = int(limit)

    try:
        while True:
            res = Feedback().list_v2(page=current_page, limit=limit, task_id=task_id, filter=filter)
            new_data = safe_get(res, ["data", "organization", "feedbackV2", "nodes"])

            if new_data is None:
                logger.warning("Warning: Expected data structure not found in API response.")
                break

            feedback_data.extend(new_data)

            current_fetched = len(new_data)
            current_page += 1

            if current_fetched <= limit:
                break
    except Exception as e:
        logger.error(f"Error fetching feedback {e}")
        if hasattr(e, "response") and hasattr(e.response, "json") and "error" in e.response.json():
            logger.error(e.response.json()["error"])
        return []

    return [_format_graphql_node(item) for item in feedback_data]



import json
import logging
import time
from typing import List, Optional

import click
import httpx

from log10._httpx_utils import _try_get
from log10.llm import Log10Config


logging.basicConfig(
    format="[%(asctime)s - %(name)s - %(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger: logging.Logger = logging.getLogger("LOG10")
logger.setLevel(logging.INFO)


_log10_config = Log10Config()


def _get_completion(id: str) -> httpx.Response:
    url = f"{_log10_config.url}/api/completions/{id}?organization_id={_log10_config.org_id}"
    return _try_get(url)


def _get_tag_id(tag: str) -> str:
    url = f"{_log10_config.url}/api/tags/search?organization_id={_log10_config.org_id}&query={tag}"
    res = _try_get(url)
    data = res.json()["data"]
    if not data:
        return []

    if len(data) == 1:
        return data[0]["id"]
    else:
        for item in data:
            if item["name"] == tag:
                return item["id"]
        return []


def _get_tag_ids(tags):
    tag_ids = []
    for tag in [t for t in tags.split(",") if t]:
        tag_id = _get_tag_id(tag)
        if tag_id:
            tag_ids.append(tag_id)
        else:
            raise SystemExit(f"Cannot found tag: {tag}.")

    tag_ids_str = ",".join(tag_ids)
    return tag_ids_str


def _get_completions_url(limit, offset, tags, from_date, to_date, base_url, org_id, printout=True):
    tag_ids_str = _get_tag_ids(tags) if tags else ""
    if tag_ids_str and printout:
        print(f"Filter with tags: {tags}")

    date_range = _get_valid_date_range(from_date, to_date)
    if date_range and printout:
        print(f"Filter with created date: {date_range['from'][:10]} to {date_range['to'][:10]}")

    url = f"{base_url}/api/completions?organization_id={org_id}&offset={offset}&limit={limit}&tagFilter={tag_ids_str}&createdFilter={json.dumps(date_range)}&sort=created_at&desc=true&ids="
    return url


def _get_valid_date_range(from_date, to_date):
    if not from_date and not to_date:
        return {}

    if from_date is None or to_date is None:
        raise click.UsageError("Both --from and --to must be set together.")

    if from_date >= to_date:
        raise click.UsageError(f"from_date {from_date} must be earlier than to_date {to_date}")

    parsed_from_date = from_date.replace(hour=8).strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"
    parsed_to_date = to_date.replace(hour=8).strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"

    date_range = {"from": parsed_from_date, "to": parsed_to_date}
    return date_range


def _get_llm_repsone(
    model: str,
    messages: list[dict],
    temperature: float = 0.2,
    max_tokens: int = 512,
    top_p: float = 1.0,
):
    ret = {"content": "", "usage": {"prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0}, "duration": 0.0}

    start_time = time.perf_counter()
    if "gpt-4" in model or "gpt-3.5" in model:
        from log10.load import OpenAI

        client = OpenAI()
        response = client.chat.completions.create(
            model=model, messages=messages, temperature=temperature, max_tokens=max_tokens, top_p=top_p
        )
        ret["content"] = response.choices[0].message.content
        ret["usage"] = response.usage.dict()
    elif "claude-3" in model:
        from log10.load import Anthropic

        system_messages = [m["content"] for m in messages if m["role"] == "system"]
        other_messages = [m for m in messages if m["role"] != "system"]
        system_prompt = ("\n").join(system_messages)

        client = Anthropic()
        response = client.messages.create(
            model=model,
            system=system_prompt,
            messages=other_messages,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
        )
        ret["content"] = response.content[0].text
        ret["usage"]["prompt_tokens"] = response.usage.input_tokens
        ret["usage"]["completion_tokens"] = response.usage.output_tokens
        ret["usage"]["total_tokens"] = response.usage.input_tokens + response.usage.output_tokens
    elif "mistral" in model:
        import mistralai
        from mistralai.client import MistralClient

        from log10.load import log10

        log10(mistralai)

        client = MistralClient()
        response = client.chat(
            model=model, messages=messages, temperature=temperature, max_tokens=max_tokens, top_p=top_p
        )
        ret["content"] = response.choices[0].message.content
        ret["usage"] = response.usage.model_dump()
    else:
        raise ValueError(f"Model {model} not supported.")
    ret["duration"] = int((time.perf_counter() - start_time) * 1000)

    return ret


def _compare(models: list[str], messages: dict, temperature: float = 0.2, max_tokens: float = 256, top_p: float = 1.0):
    ret = {}
    if models:
        for model in models:
            print(f"Running {model}")
            response = _get_llm_repsone(
                model,
                messages,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
            )

            ret[model] = response
    return ret


_SUPPORTED_MODELS = [
    # openai chat models
    "gpt-4o",
    "gpt-4o-2024-05-13",
    "gpt-4o-2024-08-06",
    "gpt-4o-mini",
    "gpt-4o-mini-2024-07-18",
    "gpt-4-turbo",
    "gpt-4-turbo-2024-04-09",
    "gpt-4-0125-preview",
    "gpt-4-turbo-preview",
    "gpt-4-1106-preview",
    "gpt-4-vision-preview",
    "gpt-4",
    "gpt-4-0314",
    "gpt-4-0613",
    "gpt-4-32k",
    "gpt-4-32k-0314",
    "gpt-4-32k-0613",
    "gpt-3.5-turbo",
    "gpt-3.5-turbo-16k",
    "gpt-3.5-turbo-0301",
    "gpt-3.5-turbo-0613",
    "gpt-3.5-turbo-1106",
    "gpt-3.5-turbo-0125",
    "gpt-3.5-turbo-16k-0613",
    # anthropic claude
    "claude-3-5-sonnet-20240620",
    "claude-3-opus-20240229",
    "claude-3-sonnet-20240229",
    "claude-3-haiku-20240307",
    # mistral
    "mistral-small-latest",
    "mistral-medium-latest",
    "mistral-large-latest",
]


def _check_model_support(model: str) -> bool:
    return model in _SUPPORTED_MODELS


class Completions:
    completions_path = "/api/completions"

    def __init__(self, log10_config: Log10Config = None):
        self._log10_config = log10_config or Log10Config()
        self._http_client = httpx.Client()
        self._http_client.headers = {
            "x-log10-token": self._log10_config.token,
            "x-log10-organization-id": self._log10_config.org_id,
            "Content-Type": "application/json",
        }

        self.org_id = self._log10_config.org_id
        self.base_url = self._log10_config.url
        self.url = f"{self.base_url}{self.completions_path}?organization_id={self.org_id}"

    def _get_completions(
        self,
        offset: int,
        limit: int,
        timeout: int,
        tag_names: Optional[List[str]] = None,
        from_date: click.DateTime = None,
        to_date: click.DateTime = None,
        printout: bool = True,
    ) -> List[dict]:
        url = _get_completions_url(limit, offset, tag_names, from_date, to_date, self.base_url, self.org_id)
        # Fetch completions
        response = _try_get(url, timeout)

        if response.status_code != 200:
            logger.error(f"Error: {response.json()}")
            return

        completions = response.json()
        return completions["data"]
